from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from pydantic import BaseModel

from recap.schemas.resource import (
    ResourceRef,
    ResourceSchema,
    ResourceTemplateRef,
    ResourceTemplateSchema,
)

if TYPE_CHECKING:
    from recap.adapter import Backend
from recap.schemas.process import (
    CampaignSchema,
    ProcessRunRef,
    ProcessRunSchema,
    ProcessTemplateRef,
    ProcessTemplateSchema,
)

try:
    from typing import Self
except ModuleNotFoundError:
    from typing import Self


SchemaT = TypeVar("SchemaT", bound=BaseModel)
ModelT = TypeVar("ModelT")


class QuerySpec(BaseModel):
    filters: dict[str, Any] = {}
    predicates: Sequence[Any] = ()
    orderings: Sequence[Any] = ()
    preloads: Sequence[str] = ()
    limit: int | None = None
    offset: int | None = None


class BaseQuery(Generic[SchemaT]):
    schema: type[SchemaT]

    def __init__(
        self: Self,
        backend: "Backend",
        *,
        model: type[SchemaT] | None = None,
        filters: dict[str, Any] | None = None,
        predicates: list[Any] | None = None,
        orderings: list[Any] | None = None,
        preloads: list[str] | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ):
        self._backend = backend
        self.model: type[SchemaT] = model or self.__class__.model  # type: ignore[attr-defined]
        self._filters = filters or {}
        self._predicates = predicates or []
        self._orderings = orderings or []
        self._preloads = preloads or []
        self._limit = limit
        self._offset = offset

    def _clone(self: Self, **overrides) -> "Self":
        """
        We can users to query data, not modify the object
        So for now we clone the sqlalchemy object and pass it
        to the user
        """
        params = dict(
            backend=self._backend,
            model=self.model,
            filters=dict(self._filters),
            predicates=list(self._predicates),
            orderings=list(self._orderings),
            preloads=list(self._preloads),
            limit=self._limit,
            offset=self._offset,
        )
        params.update(overrides)
        clone = self.__class__(**params)
        return clone

    def filter(self, **kwargs) -> "Self":
        new_filters = dict(self._filters)
        new_filters.update(kwargs)
        return self._clone(filters=new_filters)

    def where(self, *predicates) -> "Self":
        return self._clone(predicates=self._predicates + list(predicates))

    def order_by(self, *orderings) -> "Self":
        return self._clone(orderings=self._orderings + list(orderings))

    def limit(self, value: int) -> "Self":
        return self._clone(limit=value)

    def offset(self, value: int) -> "Self":
        return self._clone(offset=value)

    def include(self, relation_name) -> "Self":
        return self._clone(preloads=self._preloads + [relation_name])

    @property
    def _spec(self) -> QuerySpec:
        return QuerySpec(
            filters=self._filters,
            predicates=self._predicates,
            orderings=self._orderings,
            preloads=self._preloads,
            limit=self._limit,
            offset=self._offset,
        )

    def _execute(self) -> list[SchemaT]:
        rows = self._backend.query(self.model, self._spec)
        return rows

    def all(self) -> Sequence[SchemaT] | Sequence[BaseModel]:
        return self._execute()

    def first(self) -> SchemaT | None:
        return next(iter(self.limit(1)._execute()), None)

    def count(self) -> int:
        return self._backend.count(self.model, self._spec)


class CampaignQuery(BaseQuery[CampaignSchema]):
    model = CampaignSchema
    default_schema = None  # e.g. recap.schemas.campaign.CampaignSchema

    def include_process_runs(
        self,
    ) -> "CampaignQuery":
        return self.include("process_run")


class ProcessRunQuery(BaseQuery[ProcessRunSchema | ProcessRunRef]):
    model = ProcessRunSchema
    default_schema = None

    def __init__(
        self,
        backend: "Backend",
        *,
        expand: bool = True,
        **kwargs,
    ):
        self._expand = expand
        model = ProcessRunSchema if expand else ProcessRunRef
        super().__init__(backend, model=model, **kwargs)

    def _clone(self, **overrides) -> "ProcessRunQuery":
        params = dict(
            backend=self._backend,
            expand=self._expand,
            filters=dict(self._filters),
            predicates=list(self._predicates),
            orderings=list(self._orderings),
            preloads=list(self._preloads),
            limit=self._limit,
            offset=self._offset,
        )
        params.update(overrides)
        return self.__class__(**params)

    def include_steps(self, *, include_parameters: bool = False) -> "ProcessRunQuery":
        if include_parameters:
            self.include("steps.parameters")
        return self.include("steps")

    def include_resources(self) -> "ProcessRunQuery":
        return self.include("resources")


class ResourceQuery(BaseQuery[ResourceSchema | ResourceRef]):
    model = ResourceSchema
    default_schema = None

    def __init__(
        self,
        backend: "Backend",
        *,
        expand: bool = True,
        **kwargs,
    ):
        self._expand = expand
        model = ResourceSchema if expand else ResourceRef
        super().__init__(backend, model=model, **kwargs)

    def _clone(self, **overrides) -> "ResourceQuery":
        params = dict(
            backend=self._backend,
            expand=self._expand,
            filters=dict(self._filters),
            predicates=list(self._predicates),
            orderings=list(self._orderings),
            preloads=list(self._preloads),
            limit=self._limit,
            offset=self._offset,
        )
        params.update(overrides)
        return self.__class__(**params)

    def include_template(self) -> "ResourceQuery":
        return self.include("template")


class ResourceTemplateQuery(BaseQuery[ResourceTemplateSchema]):
    model = ResourceTemplateSchema
    default_schema = None

    def __init__(
        self,
        backend: "Backend",
        *,
        expand: bool = True,
        **kwargs,
    ):
        self._expand = expand
        model = ResourceTemplateSchema if expand else ResourceTemplateRef
        super().__init__(backend, model=model, **kwargs)

    def _clone(self, **overrides) -> "ResourceTemplateQuery":
        params = dict(
            backend=self._backend,
            expand=self._expand,
            filters=dict(self._filters),
            predicates=list(self._predicates),
            orderings=list(self._orderings),
            preloads=list(self._preloads),
            limit=self._limit,
            offset=self._offset,
        )
        params.update(overrides)
        return self.__class__(**params)

    def filter_by_types(self, type_list: list[str]) -> "ResourceTemplateQuery":
        return self.filter(types__names_in=type_list)

    def include_children(self) -> "ResourceTemplateQuery":
        return self.include("children")

    def include_attribute_groups(self) -> "ResourceTemplateQuery":
        return self.include("attribute_group_templates")

    def include_types(self) -> "ResourceTemplateQuery":
        return self.include("types")


class ProcessTemplateQuery(BaseQuery[ProcessTemplateSchema]):
    model = ProcessTemplateSchema
    default_schema = None

    def __init__(
        self,
        backend: "Backend",
        *,
        expand: bool = True,
        **kwargs,
    ):
        self._expand = expand
        model = ProcessTemplateSchema if expand else ProcessTemplateRef
        super().__init__(backend, model=model, **kwargs)

    def _clone(self, **overrides) -> "ProcessTemplateQuery":
        params = dict(
            backend=self._backend,
            expand=self._expand,
            filters=dict(self._filters),
            predicates=list(self._predicates),
            orderings=list(self._orderings),
            preloads=list(self._preloads),
            limit=self._limit,
            offset=self._offset,
        )
        params.update(overrides)
        return self.__class__(**params)

    def include_step_templates(self) -> "ProcessTemplateQuery":
        return self.include("step_templates")

    def include_resource_slots(self) -> "ProcessTemplateQuery":
        return self.include("resource_slots")


class QueryDSL:
    def __init__(self, backend: "Backend"):
        self.backend = backend

    def campaigns(self) -> CampaignQuery:
        return CampaignQuery(self.backend)

    def process_runs(self, *, expand: bool = True) -> ProcessRunQuery:
        return ProcessRunQuery(self.backend, expand=expand)

    def process_templates(self, *, expand: bool = True) -> ProcessTemplateQuery:
        return ProcessTemplateQuery(self.backend, expand=expand)

    def resources(self, *, expand: bool = True) -> ResourceQuery:
        return ResourceQuery(self.backend, expand=expand)

    def resource_templates(self, *, expand: bool = True) -> ResourceTemplateQuery:
        return ResourceTemplateQuery(self.backend, expand=expand)



from sqlalchemy import select

from recap.db.campaign import Campaign
from recap.db.step import StepTemplate
from recap.utils.database import get_or_create
from recap.utils.general import generate_uppercase_alphabets


def test_fragment_screening(db_session):
    from recap.db.attribute import (
        AttributeGroupTemplate,
        AttributeTemplate,
    )

    # noqa
    from recap.db.resource import (
        Resource,
        ResourceTemplate,
        ResourceType,
    )

    """
    Testing fragment screening

    - Create container Resource type
    - Library plate creation
        - Create a library plate template
        - Create library well templates and associate with plate template
        - Create a library plate resource
    - Xtal plate creation
        - Create an xtal plate template
        - Create xtal well type template and associate with xtal template
        - Create Xtal plate resource
    - Process Template creation
        - Create Process template
        - Create resource slots
        - Create Step templates and connect them to resource slots
    - Process Creation
        - Create Process from template
        - Add resources to the process
    """
    # Create container Resource type
    # container_type = ResourceType(name="container")
    container_type, _ = get_or_create(
        db_session, ResourceType, where={"name": "container"}
    )

    # Create a library plate template
    lib_plate_1536_template = ResourceTemplate(
        name="Library Plate 1536", types=[container_type]
    )
    plate_dimensions_attr = AttributeGroupTemplate(
        name="LB1536_dimensions",
    )

    num_rows_attr_val = AttributeTemplate(
        name="rows",
        value_type="int",
        default_value="32",
    )
    num_cols_attr_val = AttributeTemplate(
        name="columns",
        value_type="int",
        default_value="48",
    )
    plate_dimensions_attr.attribute_templates.append(num_rows_attr_val)
    plate_dimensions_attr.attribute_templates.append(num_cols_attr_val)

    lib_plate_1536_template.attribute_group_templates.append(plate_dimensions_attr)
    db_session.add(lib_plate_1536_template)
    db_session.commit()
    statement = select(ResourceTemplate).where(
        ResourceTemplate.name == "Library Plate 1536"
    )
    lib_plate_1536_template = db_session.scalars(statement).one()
    assert (
        lib_plate_1536_template.attribute_group_templates[0]
        .attribute_templates[0]
        .default_value
        == "32"
    )

    # Create library well templates and associate with plate template
    a_to_af = generate_uppercase_alphabets(32)
    lib_well_type_names_1536 = [
        {"name": f"{i}{str(j).zfill(2)}"} for i in a_to_af for j in range(1, 49)
    ]

    # Well attributes
    used = AttributeGroupTemplate(
        name="well_status",
    )
    used_value = AttributeTemplate(name="used", value_type="bool", default_value="true")
    used.attribute_templates.append(used_value)
    content_attr = AttributeGroupTemplate(
        name="content",
    )
    catalog_id = AttributeTemplate(
        name="catalog_id", value_type="str", default_value=""
    )
    smiles = AttributeTemplate(name="SMILES", value_type="str", default_value="")
    sequence = AttributeTemplate(name="sequence", value_type="int", default_value="0")

    content_attr.attribute_templates.append(catalog_id)
    content_attr.attribute_templates.append(smiles)
    content_attr.attribute_templates.append(sequence)

    for well_data in lib_well_type_names_1536:
        well = ResourceTemplate(name=well_data["name"], types=[container_type])
        well.attribute_group_templates.append(used)
        well.attribute_group_templates.append(content_attr)
        db_session.add(well)
        lib_plate_1536_template.children[well.name] = well

    db_session.commit()
    statement = select(ResourceTemplate).where(
        ResourceTemplate.name == "Library Plate 1536"
    )
    lib_plate_1536_template = db_session.scalars(statement).one()
    assert lib_plate_1536_template.children["A01"].name == "A01"

    # Create a library plate resource
    lib_plate = Resource(
        name="Test LP1536",
        template=lib_plate_1536_template,
    )
    db_session.add(lib_plate)
    db_session.commit()

    statement = select(Resource).where(Resource.name == "Test LP1536")
    lib_plate = db_session.scalars(statement).one()
    assert lib_plate.children["A01"].template.name == "A01"
    assert lib_plate.properties["LB1536_dimensions"].values["rows"] == 32

    from recap.db.attribute import (
        AttributeGroupTemplate,
        AttributeTemplate,
    )

    # noqa
    from recap.db.resource import Resource, ResourceTemplate  # noqa

    # - Create an xtal plate template
    xtal_plate_type = ResourceTemplate(name="SwissCI-MRC-2d", types=[container_type])
    a_to_h = generate_uppercase_alphabets(8)
    a_to_p = generate_uppercase_alphabets(16)

    echo = [f"{i}{j}" for i in a_to_p for j in range(1, 13)]
    shifter = [f"{i}{k}{j}" for i in a_to_h for j in ["a", "b"] for k in range(1, 13)]
    plate_maps = [
        {"echo": i, "shifter": j} for i, j in zip(echo, shifter, strict=False)
    ]

    well_position = AttributeGroupTemplate(name="well_position")
    well_pos_x = AttributeTemplate(name="x", value_type="int", default_value="0")
    well_pos_y_offset_0 = AttributeTemplate(
        name="y_0",
        value_type="int",
        default_value="0",
    )
    well_pos_y_offset_1350 = AttributeTemplate(
        name="y_1350",
        value_type="int",
        default_value="1350",
    )
    well_position.attribute_templates.append(well_pos_x)
    well_position.attribute_templates.append(well_pos_y_offset_0)
    well_position.attribute_templates.append(well_pos_y_offset_1350)

    # - Create xtal well type template and associate with xtal template
    for plate_map in plate_maps:
        echo_pos_attr = AttributeGroupTemplate(name="echo_pos")
        x_offset = well_pos_x
        if plate_map["shifter"][-1] == "b":
            y_offset = well_pos_y_offset_0
        else:
            y_offset = well_pos_y_offset_1350

        xtal_well_type = ResourceTemplate(
            name=plate_map["shifter"],
            types=[container_type],
        )
        echo_pos = AttributeTemplate(
            name=f"echo_pos_{plate_map['echo']}",
            value_type="str",
            default_value=plate_map["echo"],
        )
        xtal_well_type.attribute_group_templates.append(echo_pos_attr)
        echo_pos_attr.attribute_templates.append(x_offset)
        echo_pos_attr.attribute_templates.append(y_offset)
        echo_pos_attr.attribute_templates.append(echo_pos)
        xtal_plate_type.children[xtal_well_type.name] = xtal_well_type

    # - Create Xtal plate resource
    xtal_plate = Resource(name="TestXtalPlate", template=xtal_plate_type)
    db_session.add(xtal_plate)
    db_session.commit()

    statement = select(Resource).where(Resource.name == "TestXtalPlate")
    xtal_plate = db_session.scalars(statement).one()
    assert xtal_plate.children["A1a"].template.name == "A1a"

    # - Create Process template
    from recap.db.process import (
        ProcessRun,
        ProcessTemplate,
        ResourceSlot,
    )

    process_template = ProcessTemplate(
        name="Fragment Screening Sample Prep", version="1.0"
    )
    #     - Create resource slots
    lib_plate_resource_slot = ResourceSlot(
        process_template=process_template,
        resource_type=container_type,
        name="library_plate",
        direction="input",
    )
    xtal_plate_resource_slot = ResourceSlot(
        process_template=process_template,
        resource_type=container_type,
        name="xtal_plate",
        direction="input",
    )
    puck_tray_resource_slot = ResourceSlot(
        process_template=process_template,
        resource_type=container_type,
        name="puck_tray",
        direction="output",
    )
    process_template.resource_slots.append(lib_plate_resource_slot)
    process_template.resource_slots.append(xtal_plate_resource_slot)
    process_template.resource_slots.append(puck_tray_resource_slot)
    db_session.add(process_template)
    db_session.commit()

    statement = select(ProcessTemplate).where(
        ProcessTemplate.name == "Fragment Screening Sample Prep"
    )
    process_template: ProcessTemplate = db_session.scalars(statement).one()
    assert any(slot.name == "library_plate" for slot in process_template.resource_slots)

    puck_collection_template = ResourceTemplate(
        name="Puck collection template",
        types=[container_type],
    )
    puck_collection = Resource(
        name="PuckCollection",
        template=puck_collection_template,
    )
    db_session.add(puck_collection)
    db_session.commit()

    #     - Create Step templates and connect them to resource slots
    # Steps for fragment screening
    # 1. Image plate
    drop_volume_val = AttributeTemplate(
        name="volume", value_type="float", unit="nL", default_value=0
    )
    drop_volume_attr = AttributeGroupTemplate(
        name="drop_volume", attribute_templates=[drop_volume_val]
    )
    image_plate_template = StepTemplate(
        name="Image plate",
        process_template=process_template,
        attribute_group_templates=[drop_volume_attr],
    )
    image_plate_template.resource_slots["xtal_container"] = xtal_plate_resource_slot

    # 2. Echo transfer
    volume_transferred_val = AttributeTemplate(
        name="volume", value_type="float", unit="nL", default_value=0
    )
    volume_transferred_attr = AttributeGroupTemplate(
        name="volume_transferred", attribute_templates=[volume_transferred_val]
    )
    batch_number_value = AttributeTemplate(
        name="batch", value_type="int", default_value=0
    )
    batch_number_attr = AttributeGroupTemplate(
        name="batch_number", attribute_templates=[batch_number_value]
    )
    echo_tx_template = StepTemplate(
        name="Echo Transfer",
        process_template=process_template,
        attribute_group_templates=[volume_transferred_attr, batch_number_attr],
    )
    echo_tx_template.resource_slots["source_container"] = lib_plate_resource_slot
    echo_tx_template.resource_slots["dest_container"] = xtal_plate_resource_slot

    # 3. Harvesting
    time_departure_val = AttributeTemplate(name="departure_time", value_type="datetime")
    time_arrival_val = AttributeTemplate(name="arrival_time", value_type="datetime")
    lsdc_sample_val = AttributeTemplate(name="sample_name", value_type="str")
    comment_val = AttributeTemplate(name="comment", value_type="str")
    harvesting_status_val = AttributeTemplate(name="status", value_type="bool")
    harvesting_attr = AttributeGroupTemplate(
        name="harvesting",
        attribute_templates=[
            time_departure_val,
            time_arrival_val,
            comment_val,
            harvesting_status_val,
        ],
    )
    lsdc_attr = AttributeGroupTemplate(
        name="lsdc", attribute_templates=[lsdc_sample_val]
    )
    harvesting_step_template = StepTemplate(
        name="Harvesting",
        process_template=process_template,
        attribute_group_templates=[harvesting_attr, lsdc_attr],
    )
    harvesting_step_template.resource_slots["source_container"] = (
        xtal_plate_resource_slot
    )
    harvesting_step_template.resource_slots["dest_container"] = puck_tray_resource_slot

    db_session.add(image_plate_template)
    db_session.add(echo_tx_template)
    db_session.add(harvesting_step_template)
    db_session.commit()

    process_template.step_templates[image_plate_template.name] = image_plate_template
    process_template.step_templates[echo_tx_template.name] = echo_tx_template
    process_template.step_templates[harvesting_step_template.name] = (
        harvesting_step_template
    )
    db_session.add(process_template)
    db_session.commit()

    campaign = Campaign(name="Test campaign", proposal="111")

    #     - Create Process from template
    process_run = ProcessRun(
        name="FS1", description="Test", template=process_template, campaign=campaign
    )
    db_session.add(process_run)
    #     - Add resources to the process
    for resource in [
        (lib_plate, lib_plate_resource_slot),
        (xtal_plate, xtal_plate_resource_slot),
        (puck_collection, puck_tray_resource_slot),
    ]:
        process_run.resources[resource[1]] = resource[0]

    db_session.commit()

    statement = select(ProcessRun).where(ProcessRun.name == "FS1")
    process_run: ProcessRun = db_session.scalars(statement).one()
    assert process_run.steps["Image plate"].template.name == "Image plate"
    assert (
        process_run.steps["Image plate"].parameters["drop_volume"].values["volume"] == 0
    )
    assert (
        process_run.steps["Echo Transfer"]
        .parameters["volume_transferred"]
        .values["volume"]
        == 0
    )
    assert (
        process_run.steps["Echo Transfer"].parameters["batch_number"].values["batch"]
        == 0
    )



import typing
from typing import Any
from uuid import UUID, uuid4

from sqlalchemy import JSON, UniqueConstraint
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import Mapped, mapped_column, relationship

from recap.db.base import Base, TimestampMixin

if typing.TYPE_CHECKING:
    from recap.db.process import ProcessRun


class Campaign(TimestampMixin, Base):
    __tablename__ = "campaign"
    id: Mapped[UUID] = mapped_column(primary_key=True, default=uuid4)
    name: Mapped[str] = mapped_column(nullable=False)
    proposal: Mapped[str] = mapped_column(nullable=False)
    saf: Mapped[str] = mapped_column(nullable=True)
    meta_data: Mapped[dict[str, Any] | None] = mapped_column(
        MutableDict.as_mutable(JSON), nullable=True
    )
    process_runs: Mapped[list["ProcessRun"]] = relationship(
        "ProcessRun", back_populates="campaign"
    )
    resources = association_proxy("process_runs", "resources")
    __table_args__ = (
        UniqueConstraint("name", "proposal", name="uq_campaign_name_proposal"),
    )



from typing import TYPE_CHECKING, Any
from uuid import UUID, uuid4

from sqlalchemy import (
    JSON,
    CheckConstraint,
    ForeignKey,
    UniqueConstraint,
    event,
)
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import Mapped, mapped_column, relationship

from recap.utils.general import from_json_value, make_slug, to_json_compatible

if TYPE_CHECKING:
    from recap.db.resource import ResourceTemplate
    from recap.db.step import StepTemplate

from .base import Base, TimestampMixin


class AttributeGroupTemplate(TimestampMixin, Base):
    __tablename__ = "attribute_group_template"
    id: Mapped[UUID] = mapped_column(primary_key=True, default=uuid4)
    name: Mapped[str] = mapped_column(nullable=False)
    slug: Mapped[str | None] = mapped_column(nullable=True)
    attribute_templates: Mapped[list["AttributeTemplate"]] = relationship(
        back_populates="attribute_group_template",
    )

    resource_template_id: Mapped[UUID | None] = mapped_column(
        ForeignKey("resource_template.id", ondelete="CASCADE"),
        nullable=True,
        index=True,
    )
    step_template_id: Mapped[UUID | None] = mapped_column(
        ForeignKey("step_template.id", ondelete="CASCADE"), nullable=True, index=True
    )

    resource_template: Mapped["ResourceTemplate"] = relationship(
        "ResourceTemplate",
        back_populates="attribute_group_templates",
        foreign_keys=resource_template_id,
    )
    step_template: Mapped["StepTemplate"] = relationship(
        "StepTemplate",
        back_populates="attribute_group_templates",
        foreign_keys=step_template_id,
    )

    __table_args__ = (
        # Enforce XOR: exactly one FK must be non-null
        CheckConstraint(
            "(resource_template_id IS NOT NULL) <> (step_template_id IS NOT NULL)",
            name="ck_attr_group_exactly_one_owner",
        ),
        # Keep names unique per owner
        UniqueConstraint(
            "resource_template_id", "name", name="uq_attr_group_name_per_resource"
        ),
        UniqueConstraint(
            "step_template_id", "name", name="uq_attr_group_name_per_step"
        ),
    )


# --- Keep slug always in sync with name ---
@event.listens_for(AttributeGroupTemplate, "before_insert", propagate=True)
def _before_insert(mapper, connection, target: AttributeGroupTemplate):
    target.slug = make_slug(target.name)


@event.listens_for(AttributeGroupTemplate, "before_update", propagate=True)
def _before_update(mapper, connection, target: AttributeGroupTemplate):
    target.slug = make_slug(target.name)


class AttributeTemplate(TimestampMixin, Base):
    __tablename__ = "attribute_template"
    id: Mapped[UUID] = mapped_column(primary_key=True, default=uuid4)
    name: Mapped[str] = mapped_column(nullable=False)
    slug: Mapped[str | None] = mapped_column(nullable=True)
    value_type: Mapped[str] = mapped_column(nullable=False)
    unit: Mapped[str | None] = mapped_column(nullable=True)
    default_value: Mapped[str | None] = mapped_column(nullable=True)
    metadata_json: Mapped[dict[str, Any] | None] = mapped_column(
        "metadata", MutableDict.as_mutable(JSON), nullable=True, default=dict
    )

    attribute_group_template_id: Mapped[UUID] = mapped_column(
        ForeignKey("attribute_group_template.id")
    )
    attribute_group_template = relationship(
        AttributeGroupTemplate, back_populates="attribute_templates"
    )


# --- Keep slug always in sync with name ---
@event.listens_for(AttributeTemplate, "before_insert", propagate=True)
def _before_insert(mapper, connection, target: AttributeTemplate):
    target.slug = make_slug(target.name)


@event.listens_for(AttributeTemplate, "before_update", propagate=True)
def _before_update(mapper, connection, target: AttributeTemplate):
    target.slug = make_slug(target.name)


class AttributeValue(TimestampMixin, Base):
    __tablename__ = "attribute_value"
    id: Mapped[UUID] = mapped_column(primary_key=True, default=uuid4)

    attribute_template_id: Mapped[UUID] = mapped_column(
        ForeignKey("attribute_template.id")
    )
    template = relationship(AttributeTemplate)

    parameter_id: Mapped[UUID] = mapped_column(
        ForeignKey("parameter.id"), nullable=True
    )
    parameter = relationship("Parameter", back_populates="_values")

    property_id: Mapped[UUID] = mapped_column(ForeignKey("property.id"), nullable=True)
    property = relationship("Property", back_populates="_values")

    value_json: Mapped[Any | None] = mapped_column("value", JSON, nullable=True)
    metadata_json: Mapped[dict[str, Any] | None] = mapped_column(
        "metadata", MutableDict.as_mutable(JSON), nullable=True, default=dict
    )

    def __init__(self, *args, **kwargs):
        raw_value = kwargs.pop("value", None)
        raw_metadata = kwargs.pop("metadata", None)
        super().__init__(*args, **kwargs)
        if self.metadata_json is None:
            self.metadata_json = {}
        if raw_metadata is not None:
            self.metadata_json.update(raw_metadata)
        if raw_value is None and self.template:
            raw_value = self.template.default_value
        if raw_value is not None:
            self.set_value(raw_value)

    def set_value(self, value):
        if not self.parameter and not self.property:
            raise ValueError("Parameter or Property must be set before assigning value")

        vt = self.template.value_type
        if vt == "enum":
            choices = (self.template.metadata_json or {}).get("choices")
            if not choices:
                raise ValueError("enum attributes require metadata.choices to be set")
            if str(value) not in choices:
                raise ValueError(
                    f"{self.template.name} must be one of {', '.join(choices)}"
                )
            value = str(value)
        self.value_json = to_json_compatible(vt, value)

    @hybrid_property
    def value(self):
        if not self.template:
            return None

        vt = self.template.value_type
        return from_json_value(vt, self.value_json)

    @value.setter
    def value(self, v):
        self.set_value(v)



"""add_sentinel

Revision ID: 42c9cafa5e9f
Revises: 6f71f1cb4edb
Create Date: 2025-12-06 12:55:58.995462

"""

from uuid import UUID

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "42c9cafa5e9f"
down_revision = "6f71f1cb4edb"
branch_labels = None
depends_on = None

ROOT_ID = UUID("00000000-0000-0000-0000-000000000001")


def upgrade() -> None:
    conn = op.get_bind()
    resource_template = sa.table(
        "resource_template",
        sa.column("id", sa.Uuid()),
        sa.column("name", sa.String()),
        sa.column("slug", sa.String()),
        sa.column("parent_id", sa.Uuid()),
    )
    insert_stmt = sa.insert(resource_template).values(
        id=ROOT_ID,
        name="__root__",
        slug="__root__",
        parent_id=None,
    )
    conn.execute(insert_stmt)


def downgrade() -> None:
    conn = op.get_bind()
    resource_template = sa.table("resource_template", sa.column("id", sa.Uuid()))
    delete_stmt = sa.delete(resource_template).where(resource_template.c.id == ROOT_ID)
    conn.execute(delete_stmt)



from typing import Any
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field, create_model, model_validator

from recap.db.step import Parameter
from recap.schemas.attribute import (
    AttributeGroupTemplateSchema,
    AttributeTemplateValidator,
)
from recap.schemas.common import CommonFields, StepStatus
from recap.schemas.resource import ResourceSchema, ResourceSlotSchema
from recap.utils.dsl import AliasMixin, build_param_values_model


def _attr_metadata(vt: Any) -> dict | None:
    meta = getattr(vt, "metadata", None)
    meta_json = getattr(vt, "metadata_json", None)
    if isinstance(meta, dict):
        return meta
    elif isinstance(meta_json, dict):
        return meta_json
    else:
        return {}


class StepTemplateRef(CommonFields):
    name: str


class StepTemplateSchema(CommonFields):
    name: str
    attribute_group_templates: list[AttributeGroupTemplateSchema]
    resource_slots: dict[str, ResourceSlotSchema]


class ParameterSchema(CommonFields):
    template: AttributeGroupTemplateSchema
    # values: dict[str, AttributeTemplateSchema]
    values: BaseModel  # dict[str, Any]

    @model_validator(mode="before")
    def coerce_from_orm(cls, data):
        if isinstance(data, Parameter):
            tmpl = data.template
            tmpl_key = tuple(
                (
                    vt.name,
                    vt.slug,
                    vt.value_type,
                    _attr_metadata(vt),
                )
                for vt in tmpl.attribute_templates
            )
            values_model = build_param_values_model(tmpl.slug or tmpl.name, tmpl_key)
            raw_values = {av.template.name: av.value for av in data._values.values()}
            return {
                "id": data.id,
                "create_date": data.create_date,
                "modified_date": data.modified_date,
                "template": tmpl,
                "values": values_model.model_validate(raw_values),
            }
        if isinstance(data, dict) and isinstance(data.get("values"), dict):
            tmpl = data.get("template")
            if tmpl:
                tmpl_key = tuple(
                    (
                        vt.name,
                        vt.slug,
                        vt.value_type,
                        _attr_metadata(vt),
                    )
                    for vt in tmpl.attribute_templates
                )
                values_model = build_param_values_model(
                    tmpl.slug or tmpl.name, tmpl_key
                )
                data["values"] = values_model.model_validate(data["values"])
        return data

    @model_validator(mode="before")
    def coerce_and_reject_unknown(cls, data):
        if not isinstance(data, dict):
            return data
        template = data.get("template")
        raw_values = data.get("values") or {}
        if template and isinstance(raw_values, dict):
            tmpl_names = {a.name for a in template.attribute_templates}
            unknown = set(raw_values) - tmpl_names
            if unknown:
                raise ValueError(
                    f"Unknown parameter(s) for template {template.name}: "
                    f"{', '.join(sorted(unknown))}"
                )
            tmpl_key = tuple(
                (
                    vt.name,
                    vt.slug,
                    vt.value_type,
                    _attr_metadata(vt),
                )
                for vt in template.attribute_templates
            )
            values_model = build_param_values_model(
                template.slug or template.name, tmpl_key
            )
            data["values"] = values_model.model_validate(raw_values)
        return data

    @model_validator(mode="after")
    def validate_and_coerce_values(self) -> "ParameterSchema":
        # Build template lookup: attr name -> template schema
        tmpl_by_name = {a.name: a for a in self.template.attribute_templates}

        values_dict = (
            self.values.model_dump(by_alias=True)
            if isinstance(self.values, BaseModel)
            else dict(self.values)
        )

        # 1) no unknown keys
        unknown_keys = set(values_dict) - set(tmpl_by_name)
        if unknown_keys:
            raise ValueError(
                f"Unknown parameter(s) for template {self.template.name}: "
                f"{', '.join(sorted(unknown_keys))}"
            )

        # 2) coerce each value using your AttributeTemplateValidator
        coerced: dict[str, Any] = {}
        for name, raw_value in values_dict.items():
            attr_tmpl = tmpl_by_name[name]

            # Reuse your validator to perform type coercion & checks
            # Note: we shove `raw_value` into 'default' to leverage coerce_default()
            validator = AttributeTemplateValidator(
                name=attr_tmpl.name,
                type=attr_tmpl.value_type,
                unit=attr_tmpl.unit,
                metadata=_attr_metadata(attr_tmpl),
                default=raw_value,
            )
            coerced[name] = validator.default  # already converted by coerce_default

        self.values = self.values.__class__.model_validate(coerced)
        return self


class StepSchema(CommonFields):
    name: str
    template: StepTemplateSchema
    parameters: BaseModel | dict[str, ParameterSchema]
    state: StepStatus
    process_run_id: UUID
    parent_id: UUID | None = None
    children: list["StepSchema"] = Field(default_factory=list)
    resources: dict[str, "ResourceSchema"] = Field(default_factory=dict)

    def generate_child(self):
        return self.model_copy(deep=True, update={"id": None, "parent_id": self.id})

    @model_validator(mode="after")
    def build_parameter_model(self) -> "StepSchema":
        if isinstance(self.parameters, BaseModel):
            return self

        param_fields: dict[str, tuple] = {}
        param_values: dict[str, ParameterSchema] = {}
        for param in self.parameters.values():
            tmpl = param.template
            field_name = getattr(tmpl, "slug", None) or tmpl.name
            param_fields[field_name] = (ParameterSchema, Field(alias=tmpl.name))
            param_values[field_name] = param

        if param_fields:
            model = create_model(
                f"StepParameters_{self.template.name}",
                __base__=(AliasMixin, BaseModel),
                __config__=ConfigDict(
                    validate_assignment=True,
                    populate_by_name=True,
                    arbitrary_types_allowed=True,
                ),
                **param_fields,
            )
            self.parameters = model.model_validate(param_values)

        return self


StepSchema.model_rebuild()



from typing import Any
from uuid import UUID

from pydantic import BaseModel
from sqlalchemy.exc import NoResultFound

from recap.adapter import Backend
from recap.dsl.attribute_builder import AttributeGroupBuilder
from recap.dsl.query import QuerySpec
from recap.schemas.attribute import AttributeTemplateValidator
from recap.schemas.process import (
    CampaignSchema,
    ProcessRunSchema,
    ProcessTemplateRef,
    ProcessTemplateSchema,
)
from recap.schemas.resource import ResourceSchema, ResourceSlotSchema
from recap.schemas.step import StepSchema, StepTemplateRef
from recap.utils.dsl import lock_instance_fields
from recap.utils.general import Direction


class ProcessTemplateBuilder:
    def __init__(
        self,
        backend: Backend,
        name: str | None,
        version: str | None,
        process_template_id: UUID | None = None,
    ):
        self.backend = backend
        self._uow = None
        self._ensure_uow()
        self.name = name
        self.version = version
        self._template: ProcessTemplateRef | None = None
        self._resource_slots: dict[str, ResourceSlotSchema] = {}
        self._current_step_builder: StepTemplateBuilder | None = None
        if process_template_id is not None:
            tmpl = self.backend.get_process_template(
                name=None, version=None, id=process_template_id, expand=False
            )
            self.name = tmpl.name
            self.version = tmpl.version
            self._template = tmpl
        elif self.name is None or self.version is None:
            raise ValueError(
                "name and version are required to create a process template"
            )

    def __enter__(self):
        self._ensure_uow()
        if self._template is not None:
            self._reload_template()
            self.name = self._template.name
            self.version = self._template.version
        return self

    def __exit__(self, exc_type, exc, tb):
        if exc_type is None:
            self.save()
        else:
            if self._uow:
                self._uow.rollback()
            self._uow = None

    def save(self):
        self._ensure_uow()
        self._uow.commit()
        self._uow = None
        return self

    @property
    def template(self) -> ProcessTemplateRef:
        if not self._template:
            raise RuntimeError(
                "Call .save() first or construct template via builder methods"
            )
        return self._template

    def _ensure_template(self):
        self._ensure_uow()
        if self._template:
            return
        if self.name is None or self.version is None:
            raise ValueError(
                "name and version are required to create a process template"
            )
        self._template = self.backend.create_process_template(self.name, self.version)

    def _reload_template(self):
        self._ensure_uow()
        self._template = self.backend.get_process_template(
            self.name, self.version, expand=True
        )

    def add_resource_slot(
        self,
        name: str,
        resource_type: str,
        direction: Direction,
        create_resource_type=False,
    ) -> "ProcessTemplateBuilder":
        self._ensure_uow()
        self._ensure_template()
        self._resource_slots[name] = self.backend.add_resource_slot(
            name, resource_type, direction, self.template, create_resource_type
        )
        return self

    def add_step(
        self,
        name: str,
    ):
        self._ensure_uow()
        self._ensure_template()
        step_template = self.backend.add_step(name, self.template)
        step_template_builder = StepTemplateBuilder(
            parent=self, step_template=step_template
        )
        return step_template_builder

    def get_model(self, *, update: bool = False) -> ProcessTemplateSchema:
        """
        Return a pydantic model for the process template, optionally reloading
        from the backend first. Critical fields are locked against mutation.
        """
        self._ensure_uow()
        if update:
            self._reload_template()
        elif self._template is None:
            self._ensure_template()

        model = self.backend.get_process_template(self.name, self.version, expand=True)
        return lock_instance_fields(
            model.model_copy(deep=True),
            {"id", "create_date", "modified_date", "version"},
        )

    def set_model(self, model: ProcessTemplateSchema | ProcessTemplateRef):
        if self._template is None:
            raise RuntimeError("Template not initialized")
        if model.id != self._template.id:
            raise ValueError("ID for this ProcessTemplate does not match the builder")
        self._template = model

    def _ensure_uow(self):
        if self._uow is None:
            self._uow = self.backend.begin()
        return self._uow


class StepTemplateBuilder:
    """Scoped editor for a single step"""

    def __init__(self, parent: ProcessTemplateBuilder, step_template: StepTemplateRef):
        self.parent: ProcessTemplateBuilder = parent
        self.backend: Backend = parent.backend
        self.process_template = parent.template
        self._template = step_template
        self._bound_slots = {}

    def close_step(self) -> ProcessTemplateBuilder:
        return self.parent

    def param_group(
        self, group_name: str
    ) -> "AttributeGroupBuilder[StepTemplateBuilder]":
        attr_group_builder: AttributeGroupBuilder[StepTemplateBuilder] = (
            AttributeGroupBuilder(group_name=group_name, parent=self)
        )
        return attr_group_builder

    def bind_slot(self, role: str, slot_name: str):
        slot = self.backend.bind_slot(
            role, slot_name, self.process_template, self._template
        )
        self._bound_slots[slot.name] = slot
        return self

    def add_parameters(self, param_def: dict[str, list[dict[str, Any]]]):
        """
        Add parameters in the form of a dictionary, first level of keys represents groupts which have a list of dictionaries representing parameters
        {
            "harvest": [
                {"name":"arrival", "type": "datetime", "unit": "", default: ""},
            ]
        }
        """
        for group_key, params in param_def.items():
            agb = AttributeGroupBuilder(group_name=group_key, parent=self)
            for param in params:
                attr = AttributeTemplateValidator.model_validate(param)
                agb.add_attribute(
                    attr.name,
                    attr.type,
                    attr.unit,
                    attr.default,
                    metadata=attr.metadata,
                )
            agb.close_group()
        return self


class ProcessRunBuilder:
    def __init__(
        self,
        name: str | None,
        description: str | None,
        template_name: str | None,
        campaign: CampaignSchema | None,
        backend: Backend,
        version: str | None = None,
        process_run_id: UUID | None = None,
    ):
        self.backend = backend
        self._uow = None
        self.name = name
        self.description = description
        self.template_name = template_name
        self.version = version
        self._process_template: ProcessTemplateSchema | ProcessTemplateRef | None = None
        try:
            if process_run_id is not None:
                self._ensure_uow()
                self._process_run = self._reload_process_run(process_run_id)
                template = self._process_run.template
                self._process_template = self.backend.get_process_template(
                    template.name, template.version, expand=True
                )
                self.name = self._process_run.name
                self.description = self._process_run.description
                self.template_name = template.name
                self.version = template.version
            else:
                if (
                    name is None
                    or description is None
                    or template_name is None
                    or version is None
                ):
                    raise ValueError(
                        "name, description, template_name, and version are required to create a process run"
                    )
                if campaign is None:
                    raise ValueError("Campaign is required to create a process run")
                self._ensure_uow()
                self._process_template = self.backend.get_process_template(
                    self.template_name, self.version, expand=True
                )
                self._process_run = self.backend.create_process_run(
                    name, description, self._process_template, campaign
                )
            self._steps = list(self._process_run.steps.values())
            self._resources = {}
        except Exception as e:
            print(f"Exception occured while creating a procces run: {e}")
            raise e

    def __enter__(self):
        self._ensure_uow()
        if getattr(self, "_process_run", None) is not None:
            self._process_run = self._reload_process_run(self._process_run.id)
            template = self._process_run.template
            self._process_template = self.backend.get_process_template(
                template.name, template.version, expand=True
            )
            self.name = self._process_run.name
            self.description = self._process_run.description
            self.template_name = template.name
            self.version = template.version
            self._steps = list(self._process_run.steps.values())
        return self

    def __exit__(self, exc_type, exc, tb):
        if exc_type is None:
            self.persist()
            self.save()
        else:
            if self._uow:
                self._uow.rollback()
            self._uow = None

    def save(self):
        self._ensure_uow()
        self._uow.commit()
        self._uow = None
        return self

    def persist(self):
        self._process_run = self.backend.update_process_run(self._process_run)
        return self

    @property
    def process_run(self) -> ProcessRunSchema:
        return self._process_run

    def set_model(self, model: ProcessRunSchema):
        if self._process_run is None:
            raise RuntimeError("ProcessRun not initialized")
        if model.id != self._process_run.id:
            raise ValueError("ID for this ProcessRun does not match the builder")
        self._process_run = model

    def assign_resource(
        # self, resource_slot_name: str, resource_name: str, resource_template_name: str
        self,
        resource_slot_name: str,
        resource: ResourceSchema,
    ) -> "ProcessRunBuilder":
        self._ensure_uow()
        resource_slot = None
        for slot in self._process_template.resource_slots:
            if slot.name == resource_slot_name:
                resource_slot = slot
                break
        # resource = self.backend.get_resource(resource_name, resource_template_name)
        if resource_slot is None:
            raise NoResultFound(f"Resource slot {resource_slot_name} not found")
        self._process_run = self.backend.assign_resource(
            resource_slot, resource, self._process_run
        )
        return self

    def _check_resource_assignment(self):
        self._ensure_uow()
        self.backend.check_resource_assignment(self._process_template, self.process_run)

    @property
    def steps(self) -> list[StepSchema]:
        self._ensure_uow()
        self._check_resource_assignment()
        if self._steps is None:
            self._steps = self.backend.get_steps(self.process_run)
        return self._steps

    def get_params(
        self,
        step_name: str | None = None,
        step_schema: StepSchema | None = None,
    ) -> type[BaseModel]:
        self._ensure_uow()
        if step_name is None and step_schema is None:
            raise ValueError("Provide step_name or step_schema to get params")
        if not step_schema and step_name:
            for step in self.steps:
                if step.name == step_name:
                    step_schema = step
                    break

        if step_schema is None:
            raise NoResultFound("Step not found with name: {step_name} ")
        self._check_resource_assignment()
        return self.backend.get_params(step_schema)

    def set_params(self, filled_params: type[BaseModel]):
        self._ensure_uow()
        self.backend.set_params(filled_params)
        if getattr(self, "_process_run", None) is not None:
            # Refresh in-memory schema so subsequent persist writes current values
            self._process_run = self._reload_process_run(self._process_run.id)
        # Persist immediately so subsequent operations see updated values
        self.persist()
        return self

    def add_child_step(
        self,
        child_step: StepSchema,
    ) -> StepSchema:
        self._ensure_uow()
        if child_step.parent_id is None:
            raise ValueError(
                f"Child step {child_step.name} has no parent_id, was the step created using generate_child()?"
            )
        if child_step.process_run_id != self._process_run.id:
            raise ValueError(
                f"Child step {child_step.name} does not belong to {self._process_run.name}"
            )
        child = self.backend.add_child_step(self.process_run, child_step)
        # refresh cached steps so subsequent operations see the new child
        self._steps = None
        return child

    def get_model(self, *, update: bool = False) -> ProcessRunSchema:
        """
        Return a pydantic model representing the process run, optionally reloading
        from the backend first. Critical fields are locked against mutation.
        """
        if update:
            self._process_run = self._reload_process_run(self._process_run.id)
        model = self._process_run.model_copy(deep=True)
        return lock_instance_fields(
            model, {"id", "create_date", "modified_date", "template"}
        )

    def _ensure_uow(self):
        if self._uow is None:
            self._uow = self.backend.begin()
        return self._uow

    def _reload_process_run(self, process_run_id: UUID) -> ProcessRunSchema:
        runs = self.backend.query(
            ProcessRunSchema,
            QuerySpec(
                filters={"id": process_run_id},
                preloads=["steps", "steps.parameters", "resources"],
            ),
        )
        if not runs:
            raise ValueError(f"ProcessRun with id {process_run_id} not found")
        return runs[0]



from recap.dsl.process_builder import ProcessTemplateBuilder
from recap.dsl.resource_builder import ResourceTemplateBuilder
from recap.utils.general import Direction


def test_resource_builder_reuse_same_resource(client):
    # create once
    with ResourceTemplateBuilder(
        name="RB-Template", type_names=["container"], backend=client.backend
    ) as rtb:
        rtb.prop_group("details").add_attribute(
            "serial", "str", "", "abc"
        ).close_group()
    # build resource and mutate props, then reopen builder and mutate again
    with client.build_resource("RB-1", "RB-Template") as rb:
        rb.resource.properties["details"].values["serial"] = "xyz"
    with client.build_resource(resource_id=rb.resource.id) as rb2:
        rb2.resource.properties["details"].values["serial"] = "xyz2"

    refreshed = (
        client.query_maker().resources().filter(name="RB-1").include_template().first()
    )
    assert refreshed.properties.details.values.serial == "xyz2"


def test_resource_template_builder_reuse_same_template(client):
    with ResourceTemplateBuilder(
        name="RTB", type_names=["container"], backend=client.backend
    ) as rtb:
        rtb.prop_group("meta").add_attribute("foo", "str", "", "").close_group()
    # reopen same template by ref and add another attribute
    existing = client.query_maker().resource_templates().filter(name="RTB").first()
    with ResourceTemplateBuilder(
        name="RTB",
        type_names=["container"],
        backend=client.backend,
        resource_template_id=existing.id,
    ) as rtb2:
        rtb2.prop_group("meta").add_attribute("bar", "str", "", "").close_group()

    refreshed = client.query_maker().resource_templates().filter(name="RTB").first()
    fields = {
        a.name for a in refreshed.attribute_group_templates[0].attribute_templates
    }
    assert {"foo", "bar"} == fields


def test_process_builder_reuse_same_run(client):
    with ProcessTemplateBuilder(
        backend=client.backend, name="PTB", version="1.0"
    ) as ptb:
        ptb.add_resource_slot(
            "slot1", "container", Direction.input, create_resource_type=True
        ).add_step("S1").param_group("pg").add_attribute(
            "v", "int", "", 1
        ).close_group().close_step()

    client.create_campaign("C-reuse", "P-reuse")
    with client.build_process_run(
        name="run-reuse",
        description="desc",
        template_name="PTB",
        version="1.0",
    ) as _:
        pass

    run = (
        client.query_maker()
        .process_runs()
        .filter(name="run-reuse")
        .include_steps(include_parameters=True)
        .first()
    )
    # Create a resource that satisfies the slot and assign it
    with client.build_resource_template(
        name="ContainerRT", type_names=["container"]
    ) as _:
        pass
    container_res = client.create_resource("SlotRes", "ContainerRT")

    with client.build_process_run(process_run_id=run.id) as prb2:
        prb2.assign_resource("slot1", container_res)
        params = prb2.get_params("S1")
        params.pg.v = 5
        prb2.set_params(params)

    refreshed = (
        client.query_maker()
        .process_runs()
        .filter(name="run-reuse")
        .include_steps(include_parameters=True)
        .first()
    )
    assert refreshed.steps["S1"].parameters.pg.values.v == 5



import pytest
from sqlalchemy.orm import sessionmaker

from recap.adapter.local import LocalBackend
from recap.db.campaign import Campaign
from recap.db.process import ProcessRun, ProcessTemplate, ResourceSlot
from recap.db.resource import Resource, ResourceTemplate, ResourceType
from recap.schemas.process import ProcessRunSchema
from recap.schemas.resource import ResourceSchema, ResourceSlotSchema
from recap.utils.general import Direction


@pytest.fixture
def backend(apply_migrations, engine):
    SessionLocal = sessionmaker(bind=engine)
    backend = LocalBackend(SessionLocal)
    uow = backend.begin()
    try:
        yield backend
    finally:
        uow.rollback()
        backend.close()


def test_assign_resource_rejects_slot_from_other_template(backend):
    rt = ResourceType(name="rt")
    pt_run = ProcessTemplate(name="PT-run", version="1")
    pt_slot = ProcessTemplate(name="PT-slot", version="1")
    slot = ResourceSlot(
        name="slot-x",
        process_template=pt_slot,
        resource_type=rt,
        direction=Direction.input,
    )
    tmpl = ResourceTemplate(name="RT", types=[rt])
    res = Resource(name="R1", template=tmpl)
    camp = Campaign(name="C", proposal="p", saf=None, meta_data=None)
    run = ProcessRun(name="run", description="", template=pt_run, campaign=camp)

    backend.session.add_all([rt, pt_run, pt_slot, slot, tmpl, res, camp, run])
    backend.session.flush()

    slot_schema = ResourceSlotSchema.model_validate(slot)
    res_schema = ResourceSchema.model_validate(res, from_attributes=True)
    run_schema = ProcessRunSchema.model_validate(run, from_attributes=True)

    with pytest.raises(ValueError, match="does not belong"):
        backend.assign_resource(slot_schema, res_schema, run_schema)


def test_assign_resource_rejects_inactive_resource(backend):
    rt = ResourceType(name="rt2")
    pt = ProcessTemplate(name="PT", version="1")
    slot = ResourceSlot(
        name="slot-y", process_template=pt, resource_type=rt, direction=Direction.input
    )
    tmpl = ResourceTemplate(name="RT2", types=[rt])
    res = Resource(name="R-inactive", template=tmpl, active=False)
    camp = Campaign(name="C2", proposal="p2", saf=None, meta_data=None)
    run = ProcessRun(name="run2", description="", template=pt, campaign=camp)

    backend.session.add_all([rt, pt, slot, tmpl, res, camp, run])
    backend.session.flush()

    slot_schema = ResourceSlotSchema.model_validate(slot)
    res_schema = ResourceSchema.model_validate(res, from_attributes=True)
    run_schema = ProcessRunSchema.model_validate(run, from_attributes=True)

    with pytest.raises(ValueError, match="inactive"):
        backend.assign_resource(slot_schema, res_schema, run_schema)


def test_assign_resource_prevents_duplicate_slot_usage(backend):
    rt = ResourceType(name="rt3")
    pt = ProcessTemplate(name="PT3", version="1")
    slot = ResourceSlot(
        name="slot-z", process_template=pt, resource_type=rt, direction=Direction.input
    )
    tmpl = ResourceTemplate(name="RT3", types=[rt])
    res1 = Resource(name="R-active-1", template=tmpl)
    res2 = Resource(name="R-active-2", template=tmpl)
    camp = Campaign(name="C3", proposal="p3", saf=None, meta_data=None)
    run = ProcessRun(name="run3", description="", template=pt, campaign=camp)

    backend.session.add_all([rt, pt, slot, tmpl, res1, res2, camp, run])
    backend.session.flush()

    slot_schema = ResourceSlotSchema.model_validate(slot)
    run_schema = ProcessRunSchema.model_validate(run, from_attributes=True)

    run_schema = backend.assign_resource(
        slot_schema,
        ResourceSchema.model_validate(res1, from_attributes=True),
        run_schema,
    )

    with pytest.raises(ValueError, match="already assigned"):
        backend.assign_resource(
            slot_schema,
            ResourceSchema.model_validate(res2, from_attributes=True),
            run_schema,
        )



from itertools import product

from recap.dsl.process_builder import ProcessRunBuilder
from recap.utils.general import Direction


def test_client(client):
    with client.build_process_template("Test", "0.0.1") as ed:
        ed.add_resource_slot(
            "Input plate 1", "container", Direction.input, create_resource_type=True
        ).add_resource_slot(
            "Input plate 2", "container", Direction.input
        ).add_resource_slot(
            "Liquid transfer operator",
            "operator",
            Direction.input,
            create_resource_type=True,
        ).add_step("Transfer").bind_slot("source", "Input plate 1").bind_slot(
            "dest", "Input plate 2"
        ).bind_slot("operator", "Liquid transfer operator").param_group(
            "volume transfer"
        ).add_attribute(
            attr_name="volume", value_type="float", unit="uL", default=0.0
        ).add_attribute(
            attr_name="rate", value_type="float", unit="uL/sec", default=0.0
        ).close_group().close_step().add_step("Heat plate").bind_slot(
            "target", "Input plate 2"
        ).param_group("heat to").add_attribute(
            "temperature", "float", "degC", "0.0"
        ).close_group().close_step()

    with client.build_resource_template(
        name="96 well plate", type_names=["container", "plate"]
    ) as rt:
        rt.prop_group("dimensions").add_attribute("rows", "float", "", 8).add_attribute(
            "columns", "float", "", 12
        )
        well_cols = "ABCDEFGH"
        well_rows = [i for i in range(1, 13)]
        well_names = [f"{wn[0]}{wn[1]}" for wn in product(well_cols, well_rows)]
        for well_name in well_names:
            rt.add_child(well_name, ["container", "well"]).prop_group(
                group_name="well_data"
            ).add_attribute(
                attr_name="sample_name", value_type="str", unit="", default=""
            ).add_attribute(
                attr_name="buffer_name",
                value_type="str",
                unit="",
                default="",
            ).add_attribute(
                attr_name="volume",
                value_type="int",
                unit="uL",
                default="0",
            ).add_attribute(
                attr_name="mixing",
                value_type="str",
                unit="",
                default="",
            ).add_attribute(
                attr_name="stock",
                value_type="bool",
                unit="",
                default="False",
            ).add_attribute(
                attr_name="notes",
                value_type="str",
                unit="",
                default="",
            ).close_group().close_child()

    with client.build_resource_template(
        name="sample holder", type_names=["container", "plate"]
    ) as rt:
        rt.prop_group("dimensions").add_attribute("rows", "int", "", 2).add_attribute(
            "columns", "int", "", 9
        )
        for well_num in range(1, 19):
            rt.add_child(str(well_num), ["container", "well"]).prop_group(
                group_name="sample_holder_well_data"
            ).add_attribute(
                attr_name="sample_name",
                value_type="str",
                unit="",
                default="",
            ).add_attribute(
                attr_name="buffer_name",
                value_type="str",
                unit="",
                default="",
            ).add_attribute(
                attr_name="volume",
                value_type="float",
                unit="uL",
                default="0",
            ).close_group().close_child()

    with client.build_resource_template(
        name="robot", type_names=["robot", "liquid_transfer", "operator"]
    ) as rt:
        rt.prop_group("details").add_attribute(
            "serial_no", "str", "", "xyz"
        ).close_group()

    client.create_campaign(name="Test campaign", proposal="1")

    sample_holder = client.create_resource("Test destination plate", "sample holder")
    source_plate = client.create_resource("96 well plate", "96 well plate")
    robot = client.create_resource("LHR", "robot")

    with client.build_process_run(
        name="test_run",
        description="This is a test",
        template_name="Test",
        version="0.0.1",
    ) as run:
        run: ProcessRunBuilder
        run.assign_resource(
            resource_slot_name="Input plate 1", resource=source_plate
        ).assign_resource(
            resource_slot_name="Input plate 2", resource=sample_holder
        ).assign_resource("Liquid transfer operator", resource=robot)

        transfer_params = run.get_params("Transfer")
        transfer_params.volume_transfer.volume = 50
        transfer_params.volume_transfer.rate = 1
        run.set_params(transfer_params)

        heat_params = run.get_params("Heat plate")
        heat_params.heat_to.temperature = 100
        run.set_params(heat_params)



from collections.abc import Iterable
from pathlib import Path
from tempfile import gettempdir
from typing import Any, overload
from urllib.parse import urlparse
from uuid import UUID, uuid4

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from recap.adapter import Backend
from recap.adapter.local import LocalBackend
from recap.dsl.process_builder import ProcessRunBuilder, ProcessTemplateBuilder
from recap.dsl.query import QueryDSL
from recap.dsl.resource_builder import ResourceBuilder, ResourceTemplateBuilder
from recap.schemas.process import CampaignSchema
from recap.schemas.resource import ResourceSchema
from recap.utils.migrations import apply_migrations


class RecapClient:
    def __init__(
        self,
        url: str | None = None,
        echo: bool = False,
    ):
        self._campaign: CampaignSchema | None = None
        self.database_path: Path | None = None
        self.backend: Backend | None = None
        if url is not None:
            parsed = urlparse(url)
            if parsed.scheme in ("http", "https"):
                raise NotImplementedError("Rest api via HTTP(S) is not yet implemented")
            elif "sqlite" in parsed.scheme:
                if parsed.path and parsed.path != "/:memory:":
                    self.database_path = Path(parsed.path)
                self.engine = create_engine(url, echo=echo)
                self._sessionmaker = sessionmaker(
                    bind=self.engine, expire_on_commit=False, future=True
                )
                self.backend = LocalBackend(self._sessionmaker)
            else:
                raise ValueError(f"Unknown scheme: {parsed.scheme}")

    def close(self):
        """Close the underlying session/engine to release SQLite locks."""
        backend = getattr(self, "backend", None)
        if backend and hasattr(backend, "close"):
            backend.close()
        engine = getattr(self, "engine", None)
        if engine:
            engine.dispose()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        self.close()

    @classmethod
    def from_sqlite(
        cls, path: str | Path | None = None, echo: bool = False
    ) -> "RecapClient":
        """
        Create or upgrade a local SQLite database and return a connected client.

        If no path is provided a new database file is created under the system
        temp directory. When the file already exists, migrations are applied to
        bring it up to date.
        """
        target_path = (
            Path(path)
            if path is not None
            else Path(gettempdir()) / f"recap-{uuid4().hex}.db"
        )
        if target_path.is_dir():
            raise ValueError("Path must point to a database file, not a directory")
        target_path.parent.mkdir(parents=True, exist_ok=True)

        db_url = f"sqlite:///{target_path}"
        apply_migrations(db_url)

        client = cls(url=db_url, echo=echo)
        client.database_path = target_path
        return client

    @overload
    def build_process_template(
        self, name: str, version: str
    ) -> ProcessTemplateBuilder: ...

    @overload
    def build_process_template(
        self, *, process_template_id: UUID
    ) -> ProcessTemplateBuilder: ...

    def build_process_template(
        self,
        *args,
        process_template_id: UUID | None = None,
        **kwargs,
    ) -> ProcessTemplateBuilder:
        if self.backend is None:
            raise RuntimeError("Backend not initialized")

        if process_template_id is not None:
            if args or kwargs:
                raise TypeError(
                    "Pass either an existing process_template_id or name/version, not both"
                )
            return ProcessTemplateBuilder(
                name=None,
                version=None,
                backend=self.backend,
                process_template_id=process_template_id,
            )

        if args:
            if len(args) != 2:
                raise TypeError("Provide name and version")
            name, version = args
        else:
            try:
                name = kwargs.pop("name")
                version = kwargs.pop("version")
            except KeyError as exc:
                raise TypeError("name and version are required") from exc
            if kwargs:
                raise TypeError(f"Unexpected keyword arguments: {', '.join(kwargs)}")

        return ProcessTemplateBuilder(name=name, version=version, backend=self.backend)

    @overload
    def build_process_run(
        self, name: str, description: str, template_name: str, version: str
    ) -> ProcessRunBuilder: ...

    @overload
    def build_process_run(self, *, process_run_id: UUID) -> ProcessRunBuilder: ...

    def build_process_run(
        self,
        *args,
        process_run_id: UUID | None = None,
        **kwargs,
    ) -> ProcessRunBuilder:
        if self.backend is None:
            raise RuntimeError("Backend not initialized")

        if process_run_id is not None:
            if args or kwargs:
                raise TypeError(
                    "Pass either an existing process_run_id or name/description/template_name/version, not both"
                )
            return ProcessRunBuilder(
                name=None,
                description=None,
                template_name=None,
                campaign=self._campaign,
                backend=self.backend,
                version=None,
                process_run_id=process_run_id,
            )

        if args:
            if len(args) != 4:
                raise TypeError(
                    "Provide exactly four positional arguments: name, description, template_name, version"
                )
            name, description, template_name, version = args
        else:
            try:
                name = kwargs.pop("name")
                description = kwargs.pop("description")
                template_name = kwargs.pop("template_name")
                version = kwargs.pop("version")
            except KeyError as exc:
                raise TypeError(
                    "name, description, template_name, and version are required"
                ) from exc
            if kwargs:
                raise TypeError(f"Unexpected keyword arguments: {', '.join(kwargs)}")

        if self._campaign is None:
            raise ValueError(
                "Campaign not set, cannot create process run. Use create_campaign() or set_campaign() first"
            )

        return ProcessRunBuilder(
            name=name,
            description=description,
            template_name=template_name,
            campaign=self._campaign,
            backend=self.backend,
            version=version,
        )

    @overload
    def build_resource_template(
        self, *, name: str, type_names: list[str], version: str = "1.0"
    ) -> ResourceTemplateBuilder: ...

    @overload
    def build_resource_template(
        self, *, resource_template_id: UUID
    ) -> ResourceTemplateBuilder: ...

    def build_resource_template(
        self,
        *,
        name: str | None = None,
        type_names: list[str] | None = None,
        version: str = "1.0",
        resource_template_id: UUID | None = None,
    ):
        if self.backend is None:
            raise RuntimeError("Backend not initialized")

        if resource_template_id is not None:
            if name is not None or type_names is not None:
                raise TypeError(
                    "Pass either an existing resource_template_id or name/type_names, not both"
                )
            return ResourceTemplateBuilder(
                name=None,
                type_names=None,
                version=version,
                backend=self.backend,
                resource_template_id=resource_template_id,
            )

        if name is None or type_names is None:
            raise TypeError("name and type_names are required")

        if isinstance(type_names, str) or not isinstance(type_names, Iterable):
            raise TypeError("type_names must be a collection, not a string")
        if not all(isinstance(item, str) for item in type_names):
            raise TypeError("type_names must only contain strings")
        return ResourceTemplateBuilder(
            name=name, type_names=type_names, version=version, backend=self.backend
        )

    @overload
    def build_resource(
        self, name: str, template_name: str, template_version: str = "1.0"
    ) -> ResourceBuilder: ...

    @overload
    def build_resource(self, *, resource_id: UUID) -> ResourceBuilder: ...

    def build_resource(
        self,
        *args,
        resource_id: UUID | None = None,
        **kwargs,
    ):
        if self.backend is None:
            raise RuntimeError("Backend not initialized")

        if resource_id is not None:
            if args or kwargs:
                raise TypeError(
                    "Pass either an existing resource_id or name/template_name, not both"
                )
            return ResourceBuilder(
                name=None,
                template_name=None,
                template_version="1.0",
                backend=self.backend,
                resource_id=resource_id,
            )

        if args:
            if len(args) != 2:
                raise TypeError("Provide name and template_name")
            name, template_name = args
            template_version = "1.0"
        else:
            try:
                name = kwargs.pop("name")
                template_name = kwargs.pop("template_name")
                template_version = kwargs.pop("template_version", "1.0")
            except KeyError as exc:
                raise TypeError("name and template_name are required") from exc
            if kwargs:
                raise TypeError(f"Unexpected keyword arguments: {', '.join(kwargs)}")

        return ResourceBuilder(
            name=name,
            template_name=template_name,
            template_version=template_version,
            backend=self.backend,
        )

    def create_resource(
        self,
        name: str,
        template_name: str,
        template_version: str = "1.0",
        parent: ResourceSchema | None = None,
    ):
        return ResourceBuilder.create(
            name=name,
            template_name=template_name,
            template_version=template_version,
            backend=self.backend,
            parent=parent,
        )

    def create_campaign(
        self,
        name: str,
        proposal: str,
        saf: str | None = None,
        metadata: dict[str, Any] | None = None,
    ):
        uow = self.backend.begin()
        try:
            self._campaign = self.backend.create_campaign(name, proposal, saf, metadata)
            uow.commit()
        except Exception:
            uow.rollback()
            raise

    def set_campaign(self, id: UUID):
        uow = self.backend.begin()
        try:
            self._campaign = self.backend.set_campaign(id)
            uow.commit()
        except Exception:
            uow.rollback()
            raise

    def query_maker(self):
        return QueryDSL(self.backend)



from datetime import datetime

from pydantic import BaseModel, ConfigDict, Field, create_model
from sqlalchemy.inspection import inspect
from sqlalchemy.orm import InstrumentedAttribute
from sqlalchemy.sql import Select


def resolve_path(
    base_model,
    stmt: Select,
    path: tuple[str, ...],
    joined_paths: dict[tuple[str, ...], type],
) -> tuple[Select, InstrumentedAttribute]:
    """
    Walk a path like ("campaign", "proposal") from base_model using SQLAlchemy
    inspection. Along the way, join relationships as needed.

    Returns:
        stmt: the possibly-modified Select with joins applied
        attr: the final attribute (typically a column) on the last entity
    """
    if not path:
        raise ValueError("Empty path")

    current_entity = base_model
    mapper = inspect(current_entity)

    *rel_path, field_name = path

    # Walk relationships (if any)
    for depth, rel_name in enumerate(rel_path):
        subpath = tuple(rel_path[: depth + 1])

        if subpath in joined_paths:
            current_entity = joined_paths[subpath]
            mapper = inspect(current_entity)
            continue

        rel = mapper.relationships.get(rel_name)
        if rel is None:
            raise ValueError(
                f"{current_entity.__name__} has no relationship '{rel_name}' "
                f"(path: {'__'.join(path)})"
            )

        rel_attr: InstrumentedAttribute = getattr(current_entity, rel.key)
        target_entity = rel.mapper.class_

        # Apply the join
        stmt = stmt.join(rel_attr)

        joined_paths[subpath] = target_entity
        current_entity = target_entity
        mapper = inspect(current_entity)

    # Now resolve the final field on the last entity
    try:
        attr = getattr(current_entity, field_name)
    except AttributeError as err:
        raise ValueError(
            f"{current_entity.__name__} has no attribute '{field_name}' "
            f"(path: {'__'.join(path)})"
        ) from err

    if not isinstance(attr, InstrumentedAttribute):
        raise ValueError(
            f"Attribute '{field_name}' on {current_entity.__name__} is not a column/relationship "
            f"(path: {'__'.join(path)})"
        )

    return stmt, attr


class AliasMixin:
    """
    This is a pydantic model mixin that allows a user to access a
    field via its name in the database. For e.g. if a paramter group is
    called "Sample Temperature", its pydantic field is converted to sample_temperature
    But if the user wants to use the original string. This can be used by:
    parameter.get("Sample Temperature") or
    parameter.get("sample_temperature") if they want to use the slugified string

    """

    def get(self, alias: str):
        for name, field in self.__class__.model_fields.items():
            if alias in (field.alias, name):
                return getattr(self, name)
        raise KeyError(f"No field with alias '{alias}'")

    def set(self, alias: str, value):
        for name, field in self.__class__.model_fields.items():
            if alias in (field.alias, name):
                setattr(self, name, value)
                return
        raise KeyError(f"No field with alias '{alias}'")

    def __getitem__(self, alias: str):
        return self.get(alias)

    def __setitem__(self, alias: str, value):
        self.set(alias, value)

    def items(self):
        for name in self.__class__.model_fields:
            yield name, getattr(self, name)

    def keys(self):
        return self.__class__.model_fields.keys()

    def values(self):
        for name in self.__class__.model_fields:
            yield getattr(self, name)


def map_dtype_to_pytype(dtype: str):
    return {
        "float": float,
        "int": int,
        "str": str,
        "bool": bool,
        "datetime": datetime,
        "array": list,
        "enum": str,
    }[dtype]


def lock_instance_fields(model: BaseModel, fields: set[str]) -> BaseModel:
    """
    Return the given model instance with selected fields made read-only.
    We override __setattr__ on the instance to guard critical attributes.
    """
    locked = set(fields)
    original_setattr = model.__setattr__

    def _locked_setattr(self, name, value):
        if name in locked:
            raise TypeError(f"{name} is read-only")
        return original_setattr(name, value)

    object.__setattr__(model, "__setattr__", _locked_setattr.__get__(model))
    return model


def build_param_values_model(group_slug: str, attr_templates):
    fields: dict[str, tuple] = {}
    for name, slug, value_type, metadata in attr_templates:
        pytype = map_dtype_to_pytype(value_type)
        meta = metadata or {}
        ge = meta.get("min") if value_type in {"int", "float"} else None
        le = meta.get("max") if value_type in {"int", "float"} else None
        fields[slug] = (
            pytype | None,
            Field(default=None, alias=name, ge=ge, le=le),
        )

    return create_model(
        f"{group_slug}_values",
        **fields,
        __base__=(AliasMixin, BaseModel),
        __config__=ConfigDict(validate_assignment=True, populate_by_name=True),
    )



from typing import TYPE_CHECKING
from uuid import UUID, uuid4

from sqlalchemy import (
    Column,
    ForeignKey,
    Table,
    UniqueConstraint,
    event,
    func,
    inspect,
    select,
)
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm import Mapped, mapped_collection, mapped_column, relationship

from recap.db.process import ResourceAssignment
from recap.utils.general import make_slug

if TYPE_CHECKING:
    from recap.db.attribute import AttributeGroupTemplate

from .base import Base, TimestampMixin

# Sentinel for root ResourceTemplate
ROOT_RESOURCE_TEMPLATE_ID = UUID("00000000-0000-0000-0000-000000000001")


def _reject_new(key, _value):
    raise KeyError(
        f"{key!r} is not a valid AttributeValue for this Parameter -"
        "keys are fixed by the template"
    )


class Property(TimestampMixin, Base):
    __tablename__ = "property"
    id: Mapped[UUID] = mapped_column(primary_key=True, default=uuid4)

    resource_id: Mapped[UUID] = mapped_column(ForeignKey("resource.id"), nullable=False)
    resource: Mapped["Resource"] = relationship(back_populates="properties")

    attribute_group_template_id: Mapped[UUID] = mapped_column(
        ForeignKey("attribute_group_template.id")
    )
    template: Mapped["AttributeGroupTemplate"] = relationship("AttributeGroupTemplate")

    _values = relationship(
        "AttributeValue",
        collection_class=mapped_collection(lambda av: av.template.name),
        back_populates="property",
        cascade="all, delete-orphan",
    )

    values = association_proxy(
        "_values",
        "value",
        creator=_reject_new,
    )

    def __init__(self, *args, **kwargs):
        from .attribute import AttributeValue  # noqa

        template: AttributeGroupTemplate = kwargs.get("template")
        super().__init__(*args, **kwargs)
        for vt in template.attribute_templates:
            av = AttributeValue(template=vt, property=self)
            av.set_value(vt.default_value)


resource_template_type_association = Table(
    "resource_template_type_association",
    Base.metadata,
    Column(
        "resource_template_id", ForeignKey("resource_template.id"), primary_key=True
    ),
    Column("resource_type_id", ForeignKey("resource_type.id"), primary_key=True),
)


class ResourceTemplate(TimestampMixin, Base):
    __tablename__ = "resource_template"
    id: Mapped[UUID] = mapped_column(primary_key=True, default=uuid4)
    name: Mapped[str] = mapped_column(nullable=False)
    slug: Mapped[str] = mapped_column(nullable=True)
    version: Mapped[str] = mapped_column(nullable=False, default="1.0")

    types: Mapped[list["ResourceType"]] = relationship(
        "ResourceType",
        secondary=resource_template_type_association,
    )
    parent_id: Mapped[UUID | None] = mapped_column(
        ForeignKey("resource_template.id"), nullable=True
    )
    parent: Mapped["ResourceTemplate"] = relationship(
        "ResourceTemplate", back_populates="children", remote_side=[id]
    )

    children: Mapped[dict[str, "ResourceTemplate"]] = relationship(
        "ResourceTemplate",
        back_populates="parent",
        collection_class=mapped_collection(lambda c: c.name),
    )

    attribute_group_templates: Mapped[list["AttributeGroupTemplate"]] = relationship(
        "AttributeGroupTemplate",
        back_populates="resource_template",
        cascade="all, delete-orphan",
    )

    __table_args__ = (
        UniqueConstraint(
            "parent_id",
            "name",
            "version",
            name="uq_resource_template_parent_name_version",
        ),
    )


# --- Keep slug always in sync with name ---
@event.listens_for(ResourceTemplate, "before_insert", propagate=True)
def _before_insert_resource_template(mapper, connection, target: ResourceTemplate):
    target.slug = make_slug(target.name)
    # Ensure the root template exists and default top-level templates to it
    if target.name != "__root__" and target.parent_id is None:
        rt_table = ResourceTemplate.__table__
        root_exists = connection.scalar(
            select(rt_table.c.id).where(rt_table.c.id == ROOT_RESOURCE_TEMPLATE_ID)
        )
        if not root_exists:
            connection.execute(
                rt_table.insert().values(
                    id=ROOT_RESOURCE_TEMPLATE_ID,
                    name="__root__",
                    slug="__root__",
                    version="1.0",
                    parent_id=None,
                )
            )
        target.parent_id = ROOT_RESOURCE_TEMPLATE_ID


@event.listens_for(ResourceTemplate, "before_update", propagate=True)
def _before_update_resource_template(mapper, connection, target: ResourceTemplate):
    target.slug = make_slug(target.name)


class ResourceType(TimestampMixin, Base):
    __tablename__ = "resource_type"
    id: Mapped[UUID] = mapped_column(primary_key=True, default=uuid4)
    name: Mapped[str] = mapped_column(unique=True, nullable=False)


class Resource(TimestampMixin, Base):
    __tablename__ = "resource"
    id: Mapped[UUID] = mapped_column(primary_key=True, default=uuid4)
    name: Mapped[str] = mapped_column(nullable=False)
    slug: Mapped[str | None] = mapped_column(nullable=True, index=True)
    active: Mapped[bool] = mapped_column(nullable=False, default=True)
    resource_template_id: Mapped[UUID] = mapped_column(
        ForeignKey("resource_template.id"), nullable=True
    )
    template: Mapped["ResourceTemplate"] = relationship()
    parent_id: Mapped[UUID | None] = mapped_column(
        ForeignKey("resource.id"), nullable=True
    )
    parent: Mapped["Resource"] = relationship(
        "Resource", back_populates="children", remote_side=[id]
    )
    children: Mapped[dict[str, "Resource"]] = relationship(
        "Resource",
        back_populates="parent",
        collection_class=mapped_collection(lambda c: c.name),
    )
    properties = relationship(
        "Property",
        collection_class=mapped_collection(lambda p: p.template.name),
        back_populates="resource",
        cascade="all, delete-orphan",
    )
    assignments: Mapped[list["ResourceAssignment"]] = relationship(
        "ResourceAssignment", back_populates="resource", cascade="all, delete-orphan"
    )
    campaigns = association_proxy("assignments", "process_run.campaign")

    def __init__(
        self,
        *args,
        _init_children: bool = True,
        _visited_children: set[UUID] | None = None,
        _max_depth: int = 10,
        **kwargs,
    ):
        resource_template = kwargs.get("template")
        super().__init__(*args, **kwargs)

        if resource_template and _init_children:
            self._initialize_from_resource_template(
                resource_template, _visited_children, _max_depth
            )

    def _initialize_from_resource_template(
        self,
        resource_template: ResourceTemplate | None = None,
        visited: set[UUID] | None = None,
        max_depth: int = 10,
    ):
        """
        Automatically initialize resource from resource_template
        - Use visited to avoid using the same resource_template to prevent cycles
        - max_depth should prevent too many recursions
        - Only add properties if not present
        """
        if not resource_template:
            return

        if max_depth <= 0:
            return

        if visited is None:
            visited = set()

        if resource_template.id in visited:
            return

        visited.add(resource_template.id)
        for prop in self.template.attribute_group_templates:
            if not any(p.template.id == prop.id for name, p in self.properties.items()):
                self.properties[prop.name] = Property(template=prop)

        for child_ct in self.template.children.values():
            if child_ct.id is not None and child_ct.id in visited:
                continue
            Resource(
                name=child_ct.name,
                template=child_ct,
                parent=self,
                _visited_children=visited,
                _max_depth=max_depth - 1,
            )

    __table_args__ = (
        UniqueConstraint(
            "parent_id",
            "name",
            name="uq_resource_parent_name",
        ),
    )


# --- Keep slug always in sync with name ---
@event.listens_for(Resource, "before_insert", propagate=True)
def _before_insert(mapper, connection, target: Resource):
    target.slug = make_slug(target.name)
    # Set active as True
    if target.active is None:
        target.active = True

    # Update all other copies of the resource to active = False
    if target.active and target.name:
        tbl = Resource.__table__
        connection.execute(
            tbl.update().where(tbl.c.name == target.name).values(active=False)
        )


def _descendant_templates_cte(template_id):
    rt = ResourceTemplate.__table__
    base = select(rt.c.id).where(rt.c.id == template_id).cte(recursive=True)
    descendants = select(rt.c.id).where(rt.c.parent_id == base.c.id)
    return base.union_all(descendants)


@event.listens_for(ResourceTemplate, "before_update", propagate=True)
@event.listens_for(ResourceTemplate, "before_delete", propagate=True)
def _guard_resource_template(mapper, connection, target: ResourceTemplate):
    state = inspect(target)
    column_changes = [
        col.key
        for col in mapper.column_attrs
        if state.attrs[col.key].history.has_changes()
        and col.key not in {"modified_date"}
    ]
    if not column_changes:
        return
    cte = _descendant_templates_cte(target.id)
    res_tbl = Resource.__table__
    count_stmt = (
        select(func.count())
        .select_from(res_tbl)
        .where(res_tbl.c.resource_template_id.in_(select(cte.c.id)))
    )
    cnt = connection.scalar(count_stmt)
    if cnt and cnt > 0:
        raise ValueError(
            "Cannot modify or delete a resource template that already has resources. "
            "Create a new template version instead."
        )


@event.listens_for(Resource, "before_update", propagate=True)
def _before_update(mapper, connection, target: Resource):
    target.slug = make_slug(target.name)
    # If this resource is set to active, set others as inactive
    if target.active and target.name:
        tbl = Resource.__table__
        connection.execute(
            tbl.update()
            .where(tbl.c.name == target.name, tbl.c.id != target.id)
            .values(active=False)
        )



from typing import Any, Self
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field, create_model, model_validator

from recap.db.resource import Property
from recap.schemas.attribute import (
    AttributeGroupTemplateSchema,
    AttributeTemplateValidator,
)
from recap.schemas.common import CommonFields
from recap.utils.dsl import AliasMixin, build_param_values_model
from recap.utils.general import Direction


def _attr_metadata(vt: Any) -> dict | None:
    meta = getattr(vt, "metadata", None)
    meta_json = getattr(vt, "metadata_json", None)
    if isinstance(meta, dict):
        return meta
    elif isinstance(meta_json, dict):
        return meta_json
    else:
        return {}


class PropertySchema(CommonFields):
    template: AttributeGroupTemplateSchema
    values: BaseModel

    @model_validator(mode="before")
    def coerce_from_orm_or_dict(cls, data):
        if isinstance(data, Property):
            tmpl = data.template
            tmpl_key = tuple(
                (
                    vt.name,
                    vt.slug,
                    vt.value_type,
                    _attr_metadata(vt),
                )
                for vt in tmpl.attribute_templates
            )
            values_model = build_param_values_model(tmpl.slug or tmpl.name, tmpl_key)
            raw_values = {av.template.name: av.value for av in data._values.values()}
            return {
                "id": data.id,
                "create_date": data.create_date,
                "modified_date": data.modified_date,
                "template": tmpl,
                "values": values_model.model_validate(raw_values),
            }
        if isinstance(data, dict) and isinstance(data.get("values"), dict):
            tmpl = data.get("template")
            if tmpl:
                tmpl_names = {a.name for a in tmpl.attribute_templates}
                unknown = set(data["values"]) - tmpl_names
                if unknown:
                    raise ValueError(
                        f"Unknown property(s) for template {tmpl.name}: "
                        f"{', '.join(sorted(unknown))}"
                    )
                tmpl_key = tuple(
                    (vt.name, vt.slug, vt.value_type, _attr_metadata(vt))
                    for vt in tmpl.attribute_templates
                )
                values_model = build_param_values_model(
                    tmpl.slug or tmpl.name, tmpl_key
                )
                data["values"] = values_model.model_validate(data["values"])
        return data

    @model_validator(mode="after")
    def validate_and_coerce_values(self) -> "PropertySchema":
        tmpl_by_name = {a.name: a for a in self.template.attribute_templates}

        values_dict = (
            self.values.model_dump(by_alias=True)
            if isinstance(self.values, BaseModel)
            else dict(self.values)
        )

        unknown_keys = set(values_dict) - set(tmpl_by_name)
        if unknown_keys:
            raise ValueError(
                f"Unknown property(s) for template {self.template.name}: "
                f"{', '.join(sorted(unknown_keys))}"
            )

        coerced: dict[str, Any] = {}
        for name, raw_value in values_dict.items():
            attr_tmpl = tmpl_by_name[name]

            validator = AttributeTemplateValidator(
                name=attr_tmpl.name,
                type=attr_tmpl.value_type,
                unit=attr_tmpl.unit,
                metadata=_attr_metadata(attr_tmpl),
                default=raw_value,
            )
            coerced[name] = validator.default

        self.values = self.values.__class__.model_validate(coerced)
        return self


class ResourceTypeSchema(CommonFields):
    name: str


class ResourceTemplateRef(CommonFields):
    name: str
    slug: str | None
    version: str
    parent: Self | None = Field(default=None, exclude=True)
    types: list[ResourceTypeSchema] = Field(default_factory=list)


class ResourceTemplateSchema(CommonFields):
    name: str
    slug: "str | None"
    version: str
    types: list[ResourceTypeSchema] = Field(default_factory=list)
    parent: ResourceTemplateRef | None = Field(default=None, exclude=True)
    children: dict[str, Self] = Field(default_factory=dict)
    attribute_group_templates: list[AttributeGroupTemplateSchema]


ResourceTypeSchema.model_rebuild()


class ResourceSlotSchema(CommonFields):
    name: str
    resource_type: ResourceTypeSchema
    direction: Direction


class ResourceSchema(CommonFields):
    name: str
    template: ResourceTemplateSchema
    parent: "ResourceRef | None" = Field(default=None, exclude=True)
    children: dict[str, Self]
    properties: BaseModel | dict[str, PropertySchema]
    model_config = ConfigDict(arbitrary_types_allowed=True, from_attributes=True)

    @model_validator(mode="after")
    def build_property_model(self) -> "ResourceSchema":
        if isinstance(self.properties, BaseModel):
            return self

        prop_fields: dict[str, tuple] = {}
        prop_values: dict[str, PropertySchema] = {}
        for prop in self.properties.values():
            tmpl = prop.template
            field_name = getattr(tmpl, "slug", None) or tmpl.name
            prop_fields[field_name] = (PropertySchema, Field(alias=tmpl.name))
            prop_values[field_name] = prop

        if prop_fields:
            model = create_model(
                f"ResourceProperties_{self.template.slug or self.template.name}",
                __base__=(AliasMixin, BaseModel),
                __config__=ConfigDict(
                    validate_assignment=True,
                    populate_by_name=True,
                    arbitrary_types_allowed=True,
                ),
                **prop_fields,
            )
            self.properties = model.model_validate(prop_values)

        return self


class ResourceRef(CommonFields):
    name: str
    template: ResourceTemplateRef


class ResourceAssignmentSchema(BaseModel):
    model_config = ConfigDict(from_attributes=True, populate_by_name=True)
    slot: ResourceSlotSchema
    resource: ResourceSchema
    step_id: UUID | None = None



import pytest

from recap.adapter.local import LocalBackend
from recap.db.attribute import AttributeGroupTemplate, AttributeTemplate, AttributeValue
from recap.db.resource import Resource, ResourceTemplate
from recap.schemas.resource import ResourceTemplateSchema


def test_attribute_value_coercion_and_json_storage(db_session):
    tmpl = ResourceTemplate(name="Machine")
    group = AttributeGroupTemplate(name="Specs", resource_template=tmpl)
    attr = AttributeTemplate(
        name="Voltage",
        value_type="int",
        default_value="10",
        attribute_group_template=group,
    )
    db_session.add_all([tmpl, group, attr])
    db_session.flush()

    res = Resource(name="R1", template=tmpl)
    db_session.add(res)
    db_session.flush()

    av = res.properties["Specs"]._values["Voltage"]
    assert av.value == 10
    assert av.value_json == 10
    assert av.metadata_json == {}

    av.value = "12"
    db_session.flush()
    assert av.value == 12
    assert av.value_json == 12


def test_attribute_value_requires_target_owner(db_session):
    tmpl = ResourceTemplate(name="SpecsTemplate")
    group = AttributeGroupTemplate(name="Specs", resource_template=tmpl)
    attr = AttributeTemplate(
        name="Orphaned",
        value_type="int",
        default_value=1,
        attribute_group_template=group,
    )
    db_session.add_all([tmpl, group, attr])
    db_session.flush()

    with pytest.raises(ValueError) as excinfo:
        AttributeValue(template=attr)  # neither parameter nor property set
    assert "Parameter or Property must be set" in str(excinfo.value)


def test_attribute_value_unsupported_type_prevents_resource_init(db_session):
    tmpl = ResourceTemplate(name="Broken")
    group = AttributeGroupTemplate(name="Bad", resource_template=tmpl)
    AttributeTemplate(
        name="BadAttr",
        value_type="unsupported",
        default_value="x",
        attribute_group_template=group,
    )
    db_session.add_all([tmpl, group])
    db_session.flush()

    with pytest.raises(ValueError):
        Resource(name="BrokenRes", template=tmpl)


def test_attribute_value_serializes_datetime_to_iso(db_session):
    tmpl = ResourceTemplate(name="Timey")
    group = AttributeGroupTemplate(name="Props", resource_template=tmpl)
    attr = AttributeTemplate(
        name="When",
        value_type="datetime",
        default_value="2024-01-01T00:00:00Z",
        attribute_group_template=group,
    )
    db_session.add_all([tmpl, group, attr])
    db_session.flush()

    res = Resource(name="Timed", template=tmpl)
    db_session.add(res)
    db_session.flush()

    av = res.properties["Props"]._values["When"]
    assert isinstance(av.value_json, str)
    assert av.value_json.startswith("2024-01-01T00:00:00")
    assert av.value.year == 2024


def test_enum_attribute_value_rejects_invalid_choice(db_session):
    tmpl = ResourceTemplate(name="Enumy")
    group = AttributeGroupTemplate(name="Choices", resource_template=tmpl)
    attr = AttributeTemplate(
        name="Position",
        value_type="enum",
        default_value="u",
        metadata_json={"choices": {"u": {}, "d": {}}},
        attribute_group_template=group,
    )
    db_session.add_all([tmpl, group, attr])
    db_session.flush()

    res = Resource(name="R", template=tmpl)
    db_session.add(res)
    db_session.flush()

    av = res.properties["Choices"]._values["Position"]
    assert av.value == "u"
    with pytest.raises(ValueError):
        av.value = "x"


def test_add_attr_group_reuses_existing_group(db_session):
    tmpl = ResourceTemplate(name="RT")
    db_session.add(tmpl)
    db_session.flush()

    backend = LocalBackend(lambda: db_session)
    # Manually bind the existing session for this test
    backend._session = db_session
    ref = ResourceTemplateSchema.model_validate(tmpl)

    first = backend.add_attr_group("content", ref)
    second = backend.add_attr_group("content", ref)

    assert first.id == second.id



import pytest
from sqlalchemy.exc import IntegrityError

from recap.db.attribute import AttributeGroupTemplate, AttributeTemplate
from recap.db.campaign import Campaign
from recap.db.process import (
    Direction,
    ProcessRun,
    ProcessTemplate,
    ResourceAssignment,
    ResourceSlot,
)
from recap.db.resource import Resource, ResourceTemplate, ResourceType
from recap.db.step import StepTemplate, StepTemplateResourceSlotBinding
from recap.utils.general import make_slug


def test_attribute_group_template_slug_and_constraint(db_session):
    process_template = ProcessTemplate(name="Pipeline", version="v1")
    step_template = StepTemplate(name="Prep", process_template=process_template)
    resource_template = ResourceTemplate(name="Instrument")
    valid_group = AttributeGroupTemplate(
        name="Group One", resource_template=resource_template
    )

    db_session.add_all(
        [process_template, step_template, resource_template, valid_group]
    )
    db_session.flush()

    assert valid_group.slug == make_slug("Group One")

    invalid_group = AttributeGroupTemplate(
        name="Invalid",
        resource_template=resource_template,
        step_template=step_template,
    )
    db_session.add(invalid_group)

    with pytest.raises(IntegrityError):
        db_session.flush()
    db_session.rollback()


def test_resource_initialization_from_template(db_session):
    resource_template = ResourceTemplate(name="Parent")
    attr_group = AttributeGroupTemplate(
        name="Settings", resource_template=resource_template
    )
    AttributeTemplate(
        name="Voltage",
        value_type="int",
        default_value="5",
        attribute_group_template=attr_group,
    )
    child_template = ResourceTemplate(name="Child", parent=resource_template)
    db_session.add_all([resource_template, attr_group, child_template])
    db_session.flush()

    resource = Resource(name="Main Resource", template=resource_template)
    db_session.add(resource)
    db_session.flush()

    assert set(resource.properties.keys()) == {"Settings"}
    settings = resource.properties["Settings"]
    assert settings.values["Voltage"] == 5
    assert len({id(child) for child in resource.children.values()}) == 1
    child = next(iter(resource.children.values()))
    assert child.template is child_template
    assert child.name == child_template.name


def test_process_run_validates_resource_assignments(db_session):
    resource_type = ResourceType(name="Microscope")
    process_template = ProcessTemplate(name="Acquisition", version="v1")
    slot = ResourceSlot(
        name="instrument",
        process_template=process_template,
        resource_type=resource_type,
        direction=Direction.input,
    )
    campaign = Campaign(name="Campaign", proposal="P1", saf=None, meta_data=None)

    matching_template = ResourceTemplate(
        name="MicroscopeTemplate", types=[resource_type]
    )
    matching_resource = Resource(name="Scope A", template=matching_template)

    wrong_type = ResourceType(name="Other")
    wrong_template = ResourceTemplate(name="OtherTemplate", types=[wrong_type])
    wrong_resource = Resource(name="Other A", template=wrong_template)

    run = ProcessRun(
        name="run-1",
        description="desc",
        template=process_template,
        campaign=campaign,
    )

    db_session.add_all(
        [
            resource_type,
            process_template,
            slot,
            campaign,
            matching_template,
            matching_resource,
            wrong_type,
            wrong_template,
            wrong_resource,
            run,
        ]
    )
    db_session.flush()

    run.resources[slot] = matching_resource
    db_session.flush()
    assert run.assignments[slot].resource is matching_resource

    slot_two = ResourceSlot(
        name="instrument-2",
        process_template=process_template,
        resource_type=resource_type,
        direction=Direction.input,
    )
    db_session.add(slot_two)
    with pytest.raises(ValueError):
        run.resources[slot_two] = wrong_resource

    another_resource = Resource(name="Scope B", template=matching_template)
    db_session.add(another_resource)
    with pytest.raises(ValueError):
        run.assignments[slot] = ResourceAssignment(
            resource_slot=slot, resource=another_resource
        )


def test_step_initializes_parameters_from_template(db_session):
    process_template = ProcessTemplate(name="Prep", version="1.0")
    step_template = StepTemplate(name="Incubate", process_template=process_template)
    attr_group = AttributeGroupTemplate(name="Conditions", step_template=step_template)
    AttributeTemplate(
        name="Duration",
        value_type="int",
        default_value=30,
        attribute_group_template=attr_group,
    )
    campaign = Campaign(name="C1", proposal="P1", saf=None, meta_data=None)

    run = ProcessRun(
        name="run-001",
        description="desc",
        template=process_template,
        campaign=campaign,
    )
    db_session.add_all([process_template, step_template, attr_group, campaign, run])
    db_session.flush()

    step = run.steps["Incubate"]
    assert step.name == step_template.name
    assert set(step.parameters.keys()) == {"Conditions"}
    param = step.parameters["Conditions"]
    assert param.values["Duration"] == 30


def test_resource_slug_and_active_toggle(db_session):
    tmpl1 = ResourceTemplate(name="Machine")
    tmpl2 = ResourceTemplate(name="Machine2")
    db_session.add_all([tmpl1, tmpl2])
    db_session.flush()

    first = Resource(name="Shared Name", template=tmpl1)
    db_session.add(first)
    db_session.flush()

    assert first.slug == make_slug("Shared Name")
    assert first.active is True

    second = Resource(name="Shared Name", template=tmpl2)
    db_session.add(second)
    db_session.flush()

    # new one is active, old one was deactivated by the before_insert hook
    assert second.active is True
    db_session.refresh(first)
    assert first.active is False


def test_resource_children_respect_max_depth(db_session):
    tmpl = ResourceTemplate(name="Parent")
    ResourceTemplate(name="Child", parent=tmpl)
    db_session.add(tmpl)
    db_session.flush()

    resource = Resource(name="Root", template=tmpl, _max_depth=0)
    db_session.add(resource)
    db_session.flush()

    assert resource.children == {}


def test_property_values_reject_unknown_keys(db_session):
    tmpl = ResourceTemplate(name="Machine")
    group = AttributeGroupTemplate(name="Props", resource_template=tmpl)
    AttributeTemplate(
        name="Voltage",
        value_type="int",
        default_value=5,
        attribute_group_template=group,
    )
    db_session.add_all([tmpl, group])
    db_session.flush()

    res = Resource(name="Machine A", template=tmpl)
    db_session.add(res)
    db_session.flush()

    with pytest.raises(KeyError):
        res.properties["Props"].values["Unknown"] = 1


def test_resource_slot_name_unique_per_process_template(db_session):
    pt = ProcessTemplate(name="PT", version="1")
    slot1 = ResourceSlot(
        name="instrument",
        process_template=pt,
        resource_type=ResourceType(name="t"),
        direction=Direction.input,
    )
    slot2 = ResourceSlot(
        name="instrument",
        process_template=pt,
        resource_type=ResourceType(name="t2"),
        direction=Direction.output,
    )
    db_session.add_all([pt, slot1, slot2])
    with pytest.raises(IntegrityError):
        db_session.flush()
    db_session.rollback()


def test_step_uses_template_name_when_missing(db_session):
    pt = ProcessTemplate(name="PT2", version="1")
    st = StepTemplate(name="T1", process_template=pt)
    campaign = Campaign(name="C2", proposal="p", saf=None, meta_data=None)
    run = ProcessRun(name="run-step", description="", template=pt, campaign=campaign)
    db_session.add_all([pt, st, campaign, run])
    db_session.flush()

    # ProcessRun __init__ auto-creates steps; ensure the default naming uses the template
    auto_step = run.steps[st.name]
    assert auto_step.name == st.name


def test_step_template_binding_role_unique(db_session):
    pt = ProcessTemplate(name="PT3", version="1")
    st = StepTemplate(name="T2", process_template=pt)
    rt = ResourceType(name="slot-type")
    slot = ResourceSlot(
        name="slot",
        process_template=pt,
        resource_type=rt,
        direction=Direction.input,
    )
    db_session.add_all([pt, st, rt, slot])
    db_session.flush()

    binding1 = StepTemplateResourceSlotBinding(
        step_template_id=st.id, resource_slot_id=slot.id, role="primary"
    )
    binding2 = StepTemplateResourceSlotBinding(
        step_template_id=st.id, resource_slot_id=slot.id, role="primary"
    )
    db_session.add_all([binding1, binding2])

    with pytest.raises(IntegrityError):
        db_session.flush()
    db_session.rollback()



from datetime import datetime
from enum import Enum
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field, model_validator


class ValueType(str, Enum):
    INT = "int"
    STR = "str"
    BOOL = "bool"
    FLOAT = "float"
    DATETIME = "datetime"
    ARRAY = "array"
    ENUM = "enum"


class StepStatus(str, Enum):
    PENDING = "PENDING"
    IN_PROGRESS = "IN_PROGRESS"
    COMPLETE = "COMPLETE"


# Mapping from ValueType enum to the Python type we expect
TYPE_MAP = {
    ValueType.INT: int,
    ValueType.STR: str,
    ValueType.BOOL: bool,
    ValueType.FLOAT: float,
    ValueType.DATETIME: datetime,
    ValueType.ARRAY: list,
    ValueType.ENUM: str,
}

# DefaultValue = Union[int, float, bool, str]
DefaultValue = int | float | bool | str | datetime | list | None


class Attribute(BaseModel):
    name: str
    slug: str
    value_type: ValueType
    default_value: DefaultValue

    @model_validator(mode="after")
    def check_default_value(self):
        if not isinstance(self.default_value, TYPE_MAP[self.value_type]):
            raise ValueError(
                f"default_value must be {TYPE_MAP[self.value_type].__name__}",
                f"got {type(self.default_value).__name__} instead.",
            )
        return self


class CommonFields(BaseModel):
    id: UUID = Field(repr=False)
    create_date: datetime = Field(repr=False)
    modified_date: datetime = Field(repr=False)

    model_config = ConfigDict(from_attributes=True, populate_by_name=True)



from pathlib import Path
from tempfile import gettempdir

import pytest

from recap.client.base_client import RecapClient
from recap.schemas.process import CampaignSchema


def test_build_process_run_requires_campaign(client):
    with pytest.raises(ValueError):
        client.build_process_run("run", "desc", "tmpl", "1.0")


def test_build_resource_template_validates_type_names(apply_migrations, db_url):
    with RecapClient(url=db_url) as client:
        with pytest.raises(TypeError):
            client.build_resource_template(name="Bad", type_names="not-a-list")

        with pytest.raises(TypeError):
            client.build_resource_template(name="Bad2", type_names=["ok", 123])


def test_from_sqlite_uses_temp_dir():
    with RecapClient.from_sqlite() as client:
        assert client.database_path is not None
        assert client.database_path.exists()
        assert client.database_path.parent == Path(gettempdir())
        client.create_campaign("name", "proposal")
        assert isinstance(client._campaign, CampaignSchema)

    if client.database_path and client.database_path.exists():
        client.database_path.unlink()


def test_from_sqlite_reuses_existing_file(tmp_path):
    db_file = tmp_path / "recap.db"

    with RecapClient.from_sqlite(db_file) as client:
        client.create_campaign("name", "proposal")
        existing_id = client._campaign.id

    with RecapClient.from_sqlite(db_file) as client:
        client.set_campaign(existing_id)
        assert client._campaign.id == existing_id
        assert client.database_path == db_file



import pytest

from recap.db.process import ProcessTemplate
from recap.db.resource import ResourceTemplate
from recap.utils.general import Direction


def test_resource_template_guard_prevents_updates_when_resources_exist(client):
    with client.build_resource_template(
        name="GuardedTemplate", type_names=["sample"], version="1.0"
    ):
        pass

    with client.build_resource("Res1", "GuardedTemplate") as rb:
        resource = rb.resource

    tmpl_schema = (
        client.query_maker()
        .resource_templates()
        .filter(name="GuardedTemplate", version="1.0")
        .first()
    )

    uow = client.backend.begin()
    try:
        tmpl_model = client.backend.session.get(ResourceTemplate, tmpl_schema.id)
        tmpl_model.name = "ShouldFail"
        with pytest.raises(
            ValueError, match="resource template that already has resources"
        ):
            client.backend.session.flush()
    finally:
        uow.rollback()

    # Ensure the original resource is still intact
    fetched = (
        client.query_maker()
        .resources()
        .filter(name=resource.name)
        .include_template()
        .first()
    )
    assert fetched is not None
    assert fetched.template.name == "GuardedTemplate"


def test_process_template_guard_prevents_updates_when_runs_exist(client):
    with client.build_resource_template(
        name="ProcRes", type_names=["container"], version="1.0"
    ):
        pass
    resource = client.create_resource("ProcRes1", "ProcRes")

    client.create_campaign("Camp", "P1", "SAF1")

    with client.build_process_template(name="PT Guard", version="1.0") as pt_builder:
        pt_builder.add_resource_slot(
            "slot1",
            "container",
            direction=Direction.input,
            create_resource_type=True,
        ).add_step("step1").param_group("pg").add_attribute(
            "x", "int", "", 0
        ).close_group().close_step()

    with client.build_process_run("RunGuard", "desc", "PT Guard", "1.0") as prb:
        prb.assign_resource("slot1", resource)

    pt_schema = (
        client.query_maker()
        .process_runs()
        .filter(name="RunGuard")
        .include_steps()
        .first()
        .template
    )

    uow = client.backend.begin()
    try:
        pt_model = client.backend.session.get(ProcessTemplate, pt_schema.id)
        pt_model.name = "ShouldFail"
        with pytest.raises(ValueError, match="process template with existing runs"):
            client.backend.session.flush()
    finally:
        uow.rollback()

    unchanged = (
        client.query_maker()
        .process_runs()
        .filter(name="RunGuard")
        .include_steps()
        .first()
        .template
    )
    assert unchanged.name == "PT Guard"



from datetime import UTC, datetime, timedelta

import pytest

from recap.db.exceptions import ValidationError
from recap.utils.general import Direction


def test_platemate_via_client_child_steps(client):  # noqa
    """
    Replicate the PlateMate workflow using RecapClient, modeling harvesting as
    child steps with step-level resource assignments (well -> pin) and step
    parameters (arrival/departure/lsdc_name/harvested).
    """
    base_time = datetime(2024, 1, 1, 12, 0, tzinfo=UTC)
    lib_wells = [f"A0{i}" for i in range(1, 4)]

    # Resource templates
    with client.build_resource_template(
        name="PM Library Plate", type_names=["container", "library_plate", "plate"]
    ) as lp:
        lp.add_properties(
            {
                "dimensions": [
                    {"name": "rows", "type": "int", "default": 3},
                    {"name": "columns", "type": "int", "default": 1},
                ]
            }
        )
        for idx, well in enumerate(lib_wells, start=1):
            (
                lp.add_child(well, ["container", "well"])
                .add_properties(
                    {
                        "status": [
                            {"name": "used", "type": "bool", "default": False},
                        ],
                        "content": [
                            {"name": "catalog_id", "type": "str", "default": ""},
                            {"name": "smiles", "type": "str", "default": ""},
                            {"name": "sequence", "type": "int", "default": idx},
                        ],
                    }
                )
                .close_child()
            )

    with client.build_resource_template(
        name="PM Xtal Plate", type_names=["container", "xtal_plate", "plate"]
    ) as xt:
        xt.add_properties(
            {
                "settings": [
                    {
                        "name": "drop_volume",
                        "type": "float",
                        "unit": "nL",
                        "default": 100,
                    }
                ]
            }
        )
        mapping = [
            {"name": "A1a", "echo": "A1", "drop": "c", "origin_y": 0},
            {"name": "A1b", "echo": "A2", "drop": "lu", "origin_y": 1350},
            {"name": "A2a", "echo": "A3", "drop": "dr", "origin_y": 0},
        ]
        drop_offsets = {"c": (0, 0), "lu": (-300, 300), "dr": (300, -300)}
        for m in mapping:
            x_off, y_off = drop_offsets[m["drop"]]
            (
                xt.add_child(m["name"], ["container", "well"])
                .add_properties(
                    {
                        "mapping": [
                            {
                                "name": "echo_position",
                                "type": "str",
                                "default": m["echo"],
                            },
                            {"name": "well_origin_x", "type": "int", "default": 0},
                            {
                                "name": "well_origin_y",
                                "type": "int",
                                "default": m["origin_y"],
                            },
                        ],
                        "drop": [
                            {"name": "code", "type": "str", "default": m["drop"]},
                            {"name": "x_offset", "type": "int", "default": x_off},
                            {"name": "y_offset", "type": "int", "default": y_off},
                        ],
                    }
                )
                .close_child()
            )

    with client.build_resource_template(
        name="Puck Collection", type_names=["container", "puck_collection"]
    ) as pc:
        pc.add_child("FGZ001", ["container", "puck"])
        pc.add_child("FGZ002", ["container", "puck"])

    with client.build_resource_template(
        name="PM Puck", type_names=["container", "puck"]
    ) as pk:
        for idx in range(1, 4):
            (
                pk.add_child(f"Pin-{idx}", ["container", "pin"])
                .add_properties(
                    {
                        "mount": [
                            {"name": "position", "type": "int", "default": idx},
                            {"name": "sample_name", "type": "str", "default": ""},
                            {"name": "departure", "type": "datetime", "default": None},
                        ]
                    }
                )
                .close_child()
            )

    # Process template
    with client.build_process_template("PM Workflow", "1.0") as pt:
        (
            pt.add_resource_slot("library_plate", "library_plate", Direction.input)
            .add_resource_slot("xtal_plate", "xtal_plate", Direction.input)
            .add_resource_slot("puck_collection", "puck_collection", Direction.output)
        )
        (
            pt.add_step(name="Echo Transfer")
            .param_group("echo")
            .add_attribute("batch", "int", "", 1)
            .add_attribute("volume", "float", "nL", 25.0)
            .close_group()
            .bind_slot("source_plate", "library_plate")
            .bind_slot("dest_plate", "xtal_plate")
            .close_step()
        )
        (
            pt.add_step(name="Harvesting")
            .param_group("harvest")
            .add_attribute("arrival", "datetime", "", None)
            .add_attribute("departure", "datetime", "", None)
            .add_attribute("lsdc_name", "str", "", "")
            .add_attribute("harvested", "bool", "", False)
            .close_group()
            .bind_slot("source_plate", "xtal_plate")
            .bind_slot("dest_puck", "puck_collection")
            .close_step()
        )

    client.create_campaign("PM Campaign", "PM-001")

    # Resources
    with client.build_resource("DSI-poised", "PM Library Plate") as lib_plate:
        for well in lib_plate.resource.children.values():
            well.properties["content"].values.catalog_id = f"CAT-{well.name}"
            well.properties["content"].values.smiles = f"SMILES-{well.name}"

    xtal_plate = client.create_resource("pmtest", "PM Xtal Plate")

    # Build puck collection and ensure duplicate child creation raises
    collection_ref = client.create_resource("Test Puck Collection", "Puck Collection")
    uow = client.backend.begin()
    try:
        puck_template = client.backend.get_resource_template("PM Puck")
        for puck_name in ["FGZ001", "FGZ002"]:
            with pytest.raises(ValidationError):
                client.backend.create_resource(
                    puck_name,
                    puck_template,
                    parent_resource=collection_ref,
                )
        # Add a uniquely named puck to verify creation still works
        client.backend.create_resource(
            "FGZ003",
            puck_template,
            parent_resource=collection_ref,
        )
        uow.commit()
    except Exception:
        uow.rollback()
        raise

    def populate_children(resource_name: str, template_name: str):
        """Ensure a resource has instantiated children from its template."""
        uow_inner = client.backend.begin()
        try:
            res = client.backend.get_resource(resource_name, template_name)
            tmpl = client.backend.get_resource_template(template_name, expand=True)
            # If children already exist, skip creation
            existing_children = client.backend.get_resource(
                resource_name, template_name, expand=True
            ).children
            if existing_children:
                uow_inner.commit()
                return
            for child_tmpl in tmpl.children.values():
                client.backend.create_resource(
                    child_tmpl.name, child_tmpl, parent_resource=res
                )
            uow_inner.commit()
        except Exception:
            uow_inner.rollback()
            raise

    # Ensure wells/pins are materialized
    def seed_children(parent_name: str, parent_template: str, child_names: list[str]):
        uow_seed = client.backend.begin()
        try:
            parent_ref = client.backend.get_resource(parent_name, parent_template)
            for child_name in child_names:
                try:
                    client.backend.get_resource(child_name, child_name)
                    continue
                except Exception:
                    child_template = client.backend.get_resource_template(child_name)
                    client.backend.create_resource(
                        child_name, child_template, parent_resource=parent_ref
                    )
            uow_seed.commit()
        except Exception:
            uow_seed.rollback()
            raise

    seed_children("DSI-poised", "PM Library Plate", lib_wells)
    xtal_child_names = [m["name"] for m in mapping]
    seed_children("pmtest", "PM Xtal Plate", xtal_child_names)

    # Process run + assignments
    with client.build_process_run(
        "pm-run", "Platemate via client", "PM Workflow", "1.0"
    ) as prb:
        prb.assign_resource("library_plate", lib_plate.resource)
        prb.assign_resource("xtal_plate", xtal_plate)
        prb.assign_resource("puck_collection", collection_ref)

        echo_params = prb.get_params("Echo Transfer")
        echo_params.echo.batch = 7
        echo_params.echo.volume = 25.0
        prb.set_params(echo_params)

        # Child steps for echo transfer with step-level assignments (lib well -> xtal well)
        xtal_wells = sorted(
            client.backend.get_resource(
                "pmtest", "PM Xtal Plate", expand=True
            ).children.values(),
            key=lambda w: w.name,
        )
        lib_children = sorted(
            client.backend.get_resource(
                "DSI-poised", "PM Library Plate", expand=True
            ).children.values(),
            key=lambda w: w.name,
        )
        assert len(xtal_wells) == 3, "xtal wells not initialized"
        assert len(lib_children) == 3, "library wells not initialized"
        echo_children_created = []
        for source, dest in zip(lib_children, xtal_wells, strict=False):
            # Generate a pydantic model for the child step
            echo_transfer_step = prb.get_model().steps["Echo Transfer"].generate_child()
            # Update its values
            echo_transfer_step.parameters.echo.values.batch = 2
            echo_transfer_step.parameters.echo.values.volume = echo_params.echo.volume
            echo_transfer_step.resources["source_plate"] = source
            echo_transfer_step.resources["dest_plate"] = dest
            # Add it to the database
            prb.add_child_step(echo_transfer_step)
            echo_child = prb.add_child_step(echo_transfer_step)
            echo_children_created.append(echo_child)
        assert len(echo_children_created) == 3

        # Child steps for harvesting with step-level assignments (well -> pin inside a puck collection)
        puck_collection = client.backend.get_resource(
            "Test Puck Collection", "Puck Collection", expand=True
        )
        # Use the explicitly created puck with pins (FGZ003) for harvesting
        puck_with_pins = next(
            (p for p in puck_collection.children.values() if p.name == "FGZ003"), None
        )
        assert puck_with_pins is not None, "FGZ003 puck not initialized"
        puck_pins = sorted(puck_with_pins.children.values(), key=lambda p: p.name)
        assert len(puck_pins) >= 2, "puck pins not initialized"
        harvest_children_created = []

        for idx, dest in enumerate(xtal_wells[:2]):
            arrival = base_time + timedelta(minutes=5 + idx * 10)
            departure = arrival + timedelta(minutes=5)
            harvest_step = prb._process_run.steps["Harvesting"].generate_child()
            harvest_step.parameters.harvest.values.arrival = arrival
            harvest_step.parameters.harvest.values.departure = departure
            harvest_step.parameters.harvest.values.lsdc_name = f"mpro-{idx + 1:02d}"
            harvest_step.parameters.harvest.values.harvested = True
            harvest_step.resources["source_plate"] = dest
            harvest_step.resources["dest_puck"] = puck_pins[idx]

            harvest_child = prb.add_child_step(harvest_step)
            lib_children[idx].properties["status"].values.used = True
            harvest_children_created.append(harvest_child)
        assert len(harvest_children_created) == 2

    # Echo rows derived from echo child steps + mapping properties
    # Echo rows from creation-time pairs (source/dest) plus dest mapping
    echo_rows = []
    for source, dest in zip(lib_children, xtal_wells, strict=False):
        mapping = dest.properties["mapping"].values
        drop = dest.properties["drop"].values
        echo_rows.append(
            {
                "dest": mapping.echo_position,
                "source": source.name,
                "volume": echo_params.echo.volume,
                "x": mapping.well_origin_x + drop.x_offset,
                "y": mapping.well_origin_y + drop.y_offset,
            }
        )

    assert [r["dest"] for r in echo_rows] == ["A1", "A2", "A3"]
    assert echo_rows[1]["y"] == 1650
    assert all(r["volume"] == 25.0 for r in echo_rows)

    # Harvest manifest via child steps
    manifest = []
    summary = []
    for idx, dest in enumerate(xtal_wells[:2]):
        arrival = base_time + timedelta(minutes=5 + idx * 10)
        departure = arrival + timedelta(minutes=5)
        catalog = f"CAT-{lib_children[idx].name}"
        lib_children[idx].properties["content"].values.catalog_id = catalog
        manifest.append(
            {
                "sample": f"mpro-{idx + 1:02d}",
                "arrival": arrival,
                "departure": departure,
                "source": dest.name,
                "dest": puck_pins[idx].name,
            }
        )
        summary.append(
            {
                "sample": f"mpro-{idx + 1:02d}",
                "catalog": catalog,
                "smiles": lib_children[idx].properties["content"].values.smiles,
                "soak_min": round((departure - base_time).total_seconds() / 60, 1),
                "harvest_sec": round((departure - arrival).total_seconds(), 1),
                "dest_resource": puck_pins[idx].name,
            }
        )

    assert [m["sample"] for m in manifest] == ["mpro-01", "mpro-02"]
    assert all(isinstance(m["arrival"], datetime) for m in manifest)
    assert summary[0]["harvest_sec"] == 300.0
    assert summary[1]["catalog"] == "CAT-A02"
    assert summary[1]["dest_resource"] == "Pin-2"

    # API gaps to consider:
    # - bulk child creation in resource builders
    # - helper to map source/dest children (library wells -> xtal wells)
    # - helper to export echo CSV/harvest manifests from a process_run



from sqlalchemy import select
from sqlalchemy.exc import MultipleResultsFound, NoResultFound
from sqlalchemy.orm import Session


def get_or_create(session: Session, model, where: dict, defaults: dict | None = None):
    stmt = select(model).filter_by(**where)
    obj = session.execute(stmt).scalar_one_or_none()
    if obj:
        # update with defaults (no destructive changes)
        if defaults:
            for k, v in defaults.items():
                if getattr(obj, k, None) is None:
                    setattr(obj, k, v)
        return obj, False
    obj = model(**{**where, **(defaults or {})})
    session.add(obj)
    session.flush()
    return obj, True


def load_single(session, statement, *, label: str):
    try:
        return session.scalars(statement).one()
    except NoResultFound as exc:
        compiled = statement.compile()
        raise LookupError(
            f"{label}: no rows matched the query: {compiled} {compiled.params}"
        ) from exc
    except MultipleResultsFound as exc:
        raise LookupError(f"{label}: expected exactly one row, got multiple") from exc



from datetime import UTC, datetime
from uuid import uuid4

import pytest

from recap.schemas.attribute import (
    AttributeGroupTemplateSchema,
    AttributeTemplateSchema,
    AttributeTemplateValidator,
)
from recap.schemas.common import Attribute, ValueType
from recap.schemas.resource import PropertySchema
from recap.schemas.step import ParameterSchema


def _now():
    return datetime.now(UTC)


def _attribute_template_schema(
    name: str,
    value_type: str,
    default_value,
    unit: str | None = None,
    metadata: dict | None = None,
):
    stamp = _now()
    return AttributeTemplateSchema(
        id=uuid4(),
        create_date=stamp,
        modified_date=stamp,
        name=name,
        slug=name.lower(),
        value_type=value_type,
        unit=unit,
        default_value=default_value,
        metadata=metadata or {},
    )


def _attribute_group_schema(name: str, templates: list[AttributeTemplateSchema]):
    stamp = _now()
    return AttributeGroupTemplateSchema(
        id=uuid4(),
        create_date=stamp,
        modified_date=stamp,
        name=name,
        slug=name.lower(),
        attribute_templates=templates,
    )


def test_attribute_schema_validates_default_value_type():
    Attribute(
        name="count",
        slug="count",
        value_type=ValueType.INT,
        default_value=1,
    )
    with pytest.raises(ValueError):
        Attribute(
            name="count",
            slug="count",
            value_type=ValueType.INT,
            default_value="ten",
        )


def test_attribute_template_validator_coerces_and_rejects_defaults():
    validator = AttributeTemplateValidator(name="Voltage", type="int", default="7")
    assert validator.default == 7

    with pytest.raises(ValueError):
        AttributeTemplateValidator(name="When", type="datetime", default="not-a-date")


def test_parameter_schema_coerces_values_and_rejects_unknown():
    voltage = _attribute_template_schema("Voltage", "int", 5)
    enabled = _attribute_template_schema("Enabled", "bool", False)
    group_schema = _attribute_group_schema("Inputs", [voltage, enabled])
    stamp = _now()
    schema = ParameterSchema(
        id=uuid4(),
        create_date=stamp,
        modified_date=stamp,
        template=group_schema,
        values={"Voltage": "10", "Enabled": "true"},
    )

    assert schema.values.voltage == 10
    assert schema.values.enabled is True

    with pytest.raises(ValueError):
        ParameterSchema(
            id=uuid4(),
            create_date=stamp,
            modified_date=stamp,
            template=group_schema,
            values={"Voltage": "5", "Unknown": 1},
        )


def test_parameter_schema_exposes_typed_values_model():
    voltage = _attribute_template_schema("Voltage", "int", 5)
    enabled = _attribute_template_schema("Enabled", "bool", False)
    group_schema = _attribute_group_schema("Inputs", [voltage, enabled])
    stamp = _now()

    schema = ParameterSchema(
        id=uuid4(),
        create_date=stamp,
        modified_date=stamp,
        template=group_schema,
        values={"Voltage": "10", "Enabled": "true"},
    )

    assert schema.values.voltage == 10
    assert schema.values.enabled is True
    assert schema.values.model_dump(by_alias=True) == {"Voltage": 10, "Enabled": True}


def test_property_schema_value_coercion_matches_template():
    temperature = _attribute_template_schema("Temp", "float", 37.5)
    group_schema = _attribute_group_schema("Environment", [temperature])
    stamp = _now()
    property_schema = PropertySchema(
        id=uuid4(),
        create_date=stamp,
        modified_date=stamp,
        template=group_schema,
        values={"Temp": "25.5"},
    )

    assert property_schema.values.temp == pytest.approx(25.5)


def test_parameter_schema_rejects_uncoercible_values():
    duration = _attribute_template_schema("Duration", "int", 0)
    group_schema = _attribute_group_schema("Timing", [duration])
    stamp = _now()
    with pytest.raises(ValueError):
        ParameterSchema(
            id=uuid4(),
            create_date=stamp,
            modified_date=stamp,
            template=group_schema,
            values={"Duration": "not-an-int"},
        )


def test_parameter_schema_respects_metadata_bounds():
    voltage = _attribute_template_schema(
        "Voltage", "int", 5, metadata={"min": 0, "max": 10}
    )
    group_schema = _attribute_group_schema("Inputs", [voltage])
    stamp = _now()

    with pytest.raises(ValueError):
        ParameterSchema(
            id=uuid4(),
            create_date=stamp,
            modified_date=stamp,
            template=group_schema,
            values={"Voltage": 11},
        )


def test_enum_attribute_validates_choices():
    drop = _attribute_template_schema(
        "Drop Position",
        "enum",
        "u",
        metadata={"choices": {"u": {"x": 0, "y": 1}, "d": {"x": 0, "y": -1}}},
    )
    group_schema = _attribute_group_schema("Positions", [drop])
    stamp = _now()

    schema = ParameterSchema(
        id=uuid4(),
        create_date=stamp,
        modified_date=stamp,
        template=group_schema,
        values={"Drop Position": "d"},
    )
    assert schema.values.get("Drop Position") == "d"

    with pytest.raises(ValueError):
        ParameterSchema(
            id=uuid4(),
            create_date=stamp,
            modified_date=stamp,
            template=group_schema,
            values={"Drop Position": "x"},
        )



import typing
from collections import namedtuple
from uuid import UUID, uuid4

from sqlalchemy import Enum, ForeignKey, UniqueConstraint, event, func, inspect, select
from sqlalchemy.ext import associationproxy
from sqlalchemy.orm import (
    Mapped,
    attribute_mapped_collection,
    mapped_collection,
    mapped_column,
    object_session,
    relationship,
    validates,
)

from recap.db.campaign import Campaign
from recap.db.step import Step, StepTemplate, StepTemplateEdge
from recap.exceptions import DuplicateResourceError
from recap.utils.general import Direction

from .base import Base, TimestampMixin

if typing.TYPE_CHECKING:
    from recap.db.resource import Resource, ResourceType

AssignedResource = namedtuple("AssignedResource", ["slot", "resource"])


class ProcessTemplate(TimestampMixin, Base):
    __tablename__ = "process_template"
    id: Mapped[UUID] = mapped_column(primary_key=True, default=uuid4)
    name: Mapped[str] = mapped_column(unique=True, nullable=False)
    version: Mapped[str] = mapped_column(nullable=False)
    is_active: Mapped[bool] = mapped_column(nullable=False, default=False)
    step_templates: Mapped[dict[str, "StepTemplate"]] = relationship(
        back_populates="process_template",
        collection_class=mapped_collection(lambda st: st.name),
    )
    edges: Mapped["StepTemplateEdge"] = relationship(
        "StepTemplateEdge",
        primaryjoin=id == StepTemplateEdge.process_template_id,
        cascade="all, delete-orphan",
    )
    resource_slots: Mapped[list["ResourceSlot"]] = relationship(
        "ResourceSlot", back_populates="process_template"
    )
    __table_args__ = (
        UniqueConstraint("name", "version", name="uq_process_template_name_version"),
    )


class ResourceSlot(TimestampMixin, Base):
    __tablename__ = "resource_slot"
    id: Mapped[UUID] = mapped_column(primary_key=True, default=uuid4)
    name: Mapped[str] = mapped_column()
    process_template_id: Mapped[UUID] = mapped_column(
        ForeignKey("process_template.id"), nullable=False
    )
    process_template: Mapped[ProcessTemplate] = relationship(
        ProcessTemplate, back_populates="resource_slots"
    )
    resource_type_id: Mapped[UUID] = mapped_column(
        ForeignKey("resource_type.id"), nullable=False
    )
    resource_type: Mapped["ResourceType"] = relationship("ResourceType")
    direction: Mapped[Direction] = mapped_column(
        Enum(Direction, name="direction_enum"), nullable=False
    )

    __table_args__ = (
        UniqueConstraint(
            "process_template_id", "name", name="uq_process_template_name"
        ),
    )


class ProcessRun(TimestampMixin, Base):
    __tablename__ = "process_run"
    id: Mapped[UUID] = mapped_column(primary_key=True, default=uuid4)

    name: Mapped[str] = mapped_column(unique=True, nullable=False)
    description: Mapped[str] = mapped_column(unique=False, nullable=False)

    process_template_id: Mapped[UUID] = mapped_column(
        ForeignKey("process_template.id"), nullable=False
    )
    template: Mapped[ProcessTemplate] = relationship()

    assignments: Mapped[dict["ResourceSlot", "ResourceAssignment"]] = relationship(
        "ResourceAssignment",
        primaryjoin="and_(ProcessRun.id==ResourceAssignment.process_run_id, ResourceAssignment.step_id==None)",
        back_populates="process_run",
        cascade="all, delete-orphan",
        collection_class=attribute_mapped_collection("resource_slot"),
    )
    resources = associationproxy.association_proxy(
        "assignments",
        "resource",
        creator=lambda res_slot, resource: ResourceAssignment(
            resource_slot=res_slot, resource=resource
        ),
    )
    steps: Mapped[dict[str, "Step"]] = relationship(
        back_populates="process_run",
        collection_class=mapped_collection(lambda s: s.name),
    )
    campaign_id: Mapped[UUID] = mapped_column(ForeignKey("campaign.id"), nullable=False)
    campaign: Mapped[Campaign] = relationship("Campaign", back_populates="process_runs")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        template: ProcessTemplate | None = kwargs.get("template")
        if template is None:
            raise ValueError("Missing template for ProcessRun")
        for step_template in template.step_templates.values():
            step = Step(template=step_template, name=step_template.name)
            # Attach via the mapped collection to ensure the dict key is derived
            # from the final name instead of a default/None during construction.
            self.steps[step.name] = step

    @validates("assignments")
    def _check_assignment(self, key, assignment: "ResourceAssignment"):
        if assignment.process_run is None:
            assignment.process_run = self
        sess = object_session(self)
        if sess is not None and assignment not in sess:
            sess.add(assignment)
        slot = assignment.resource_slot
        resource = assignment.resource

        # Resource must advertise the slot's type via its template's types
        resource_type_ids = {rt.id for rt in resource.template.types}
        if slot.resource_type_id not in resource_type_ids:
            raise ValueError(
                f"Resource {resource.name} does not match required type for slot {slot.name}"
            )

        # Slot must not already be used in this run
        for existing in self.assignments.values():
            if existing is assignment:
                continue
            if existing.resource_slot_id == slot.id:
                raise ValueError(
                    f"Slot {slot.name} is already occupied in run {self.id}"
                )

        return assignment

    @property
    def assigned_resources(self):
        ar = []
        for resource_slot, resource in self.resources.items():
            ar.append(AssignedResource(slot=resource_slot, resource=resource))
        return ar


@event.listens_for(ProcessTemplate, "before_update", propagate=True)
@event.listens_for(ProcessTemplate, "before_delete", propagate=True)
def _guard_process_template(mapper, connection, target: ProcessTemplate):
    state = inspect(target)
    column_changes = [
        col.key
        for col in mapper.column_attrs
        if state.attrs[col.key].history.has_changes()
        and col.key not in {"modified_date"}
    ]
    if not column_changes:
        return
    count_stmt = (
        select(func.count())
        .select_from(ProcessRun)
        .where(ProcessRun.process_template_id == target.id)
    )
    cnt = connection.scalar(count_stmt)
    if cnt and cnt > 0:
        raise ValueError(
            "Cannot modify or delete a process template with existing runs. "
            "Create a new template version instead."
        )


class ResourceAssignment(TimestampMixin, Base):
    __tablename__ = "resource_assignment"
    id: Mapped[UUID] = mapped_column(primary_key=True, default=uuid4)
    process_run_id: Mapped[UUID] = mapped_column(
        ForeignKey("process_run.id"), nullable=False
    )
    resource_slot_id: Mapped[UUID] = mapped_column(
        ForeignKey("resource_slot.id"), nullable=False
    )
    step_id: Mapped[UUID | None] = mapped_column(ForeignKey("step.id"), nullable=True)
    resource_id: Mapped[UUID] = mapped_column(ForeignKey("resource.id"), nullable=False)

    process_run: Mapped["ProcessRun"] = relationship("ProcessRun")
    resource_slot: Mapped["ResourceSlot"] = relationship()
    resource: Mapped["Resource"] = relationship(
        "Resource", back_populates="assignments"
    )
    step: Mapped["Step | None"] = relationship("Step", back_populates="assignments")

    @validates("resource")
    def _check_resource_campaign_uniqueness(self, key, resource: "Resource"):
        if self.process_run and self.process_run.campaign:
            campaign_id = self.process_run.campaign.id
            for assignment in resource.assignments:
                if assignment is self:
                    continue
                if (
                    assignment.process_run
                    and assignment.process_run.campaign_id == campaign_id
                    and assignment.resource.parent_id == resource.parent_id
                    and assignment.step_id == self.step_id
                ):
                    raise DuplicateResourceError(
                        resource.name,
                        assignment.process_run.campaign.name,
                        assignment.process_run.name,
                        assignment.step.name if assignment.step else None,
                    )
        return resource

    __table_args__ = (
        UniqueConstraint(
            "process_run_id", "resource_slot_id", "step_id", name="uq_run_slot_step"
        ),
    )



from __future__ import annotations

from uuid import uuid4

from sqlalchemy.orm import sessionmaker

from recap.adapter.local import LocalBackend
from recap.db.attribute import AttributeGroupTemplate, AttributeTemplate
from recap.db.campaign import Campaign
from recap.db.process import ProcessRun, ProcessTemplate, ResourceSlot
from recap.db.resource import Resource, ResourceTemplate, ResourceType
from recap.db.step import StepTemplate
from recap.dsl.query import QueryDSL
from recap.schemas.process import ProcessRunRef, ProcessTemplateRef
from recap.schemas.resource import ResourceRef, ResourceTemplateRef
from recap.utils.general import Direction


def make_query(db_session):
    SessionLocal = sessionmaker(bind=db_session.get_bind())
    backend = LocalBackend(SessionLocal)
    return QueryDSL(backend)


def seed_process_run(
    db_session,
    *,
    name: str,
    with_parameters: bool = False,
    with_resource: bool = False,
) -> tuple[Campaign, ProcessRun]:
    with db_session.no_autoflush:
        campaign = Campaign(name=f"Campaign-{name}", proposal=f"PROP-{name}")
        template = ProcessTemplate(name=f"Template-{name}", version="1.0")
        step_template = StepTemplate(name=f"Step-{name}", process_template=template)

        if with_parameters:
            attr = AttributeGroupTemplate(name=f"Exposure-{name}")
            attr.attribute_templates.append(
                AttributeTemplate(
                    name="dwell_time", value_type="int", default_value="5"
                )
            )
            step_template.attribute_group_templates.append(attr)
        run = ProcessRun(
            name=f"Run-{name}",
            description=f"Process run for {name}",
            template=template,
            campaign=campaign,
        )

    db_session.add_all([campaign, template, step_template])
    db_session.flush()

    db_session.add(run)
    db_session.flush()

    if with_resource:
        resource_type = ResourceType(name=f"resource-type-{uuid4().hex}")
        resource_template = ResourceTemplate(name=f"resource-template-{uuid4().hex}")
        resource_template.types.append(resource_type)
        slot = ResourceSlot(
            name=f"slot-{name}",
            process_template=template,
            resource_type=resource_type,
            direction=Direction.input,
        )
        resource = Resource(name=f"Resource-{name}", template=resource_template)
        db_session.add_all([resource_type, resource_template, slot, resource])
        db_session.flush()
        run.resources[slot] = resource

    db_session.commit()
    return campaign, run


def test_campaign_without_include_requires_lazy_load(db_session):
    campaign, _ = seed_process_run(db_session, name="lazy")
    campaign_row = make_query(db_session).campaigns().filter(id=campaign.id).first()
    assert campaign_row is not None
    # with pytest.raises(DetachedInstanceError):
    _ = len(campaign_row.process_runs)


def test_campaign_include_process_runs_and_steps(db_session):
    campaign, _ = seed_process_run(db_session, name="include", with_parameters=True)

    loaded_campaign = (
        make_query(db_session)
        .campaigns()
        .filter(id=campaign.id)
        .include_process_runs()  # lambda q: q.include_steps(include_parameters=True))
        .first()
    )
    assert loaded_campaign is not None
    assert loaded_campaign.process_runs[0].name.startswith("Run-include")
    step = loaded_campaign.process_runs[0].steps["Step-include"]
    exposure = step.parameters["Exposure-include"]
    assert exposure.values.dwell_time == 5


def test_process_run_pagination_and_filtering(db_session):
    runs = [seed_process_run(db_session, name=f"batch-{idx}")[1] for idx in range(3)]
    names = sorted(run.name for run in runs)

    query = (
        make_query(db_session)
        .process_runs()
        .where(ProcessRun.name.like("Run-batch%"))
        .order_by(ProcessRun.name)
    )

    # head = query.limit(2).as_models()
    # assert [run.name for run in head] == names[:2]

    third = query.offset(2).first()
    assert third.name == names[2]

    filtered = query.where(ProcessRun.name == names[1]).all()
    assert [run.name for run in filtered] == [names[1]]


def test_process_run_include_resources(db_session):
    _, run = seed_process_run(db_session, name="resources", with_resource=True)

    loaded_run = (
        make_query(db_session)
        .process_runs()
        .filter(id=run.id)
        .include_resources()
        .first()
    )

    assert loaded_run is not None
    assignment = next(iter(loaded_run.assigned_resources))
    assert assignment.resource.name.startswith("Resource-resources")


def test_process_run_query_can_return_ref(db_session):
    _, run = seed_process_run(db_session, name="ref-run")

    ref = make_query(db_session).process_runs(expand=False).filter(id=run.id).first()

    assert isinstance(ref, ProcessRunRef)
    assert isinstance(ref.template, ProcessTemplateRef)
    # Ref objects should not expose steps
    assert not hasattr(ref, "steps")


def test_process_template_query_can_return_ref(db_session):
    _, run = seed_process_run(db_session, name="pt-ref")

    ref = (
        make_query(db_session)
        .process_templates(expand=False)
        .filter(id=run.template.id)
        .first()
    )

    assert isinstance(ref, ProcessTemplateRef)
    assert not hasattr(ref, "step_templates")


def test_process_template_includes(db_session):
    _, run = seed_process_run(db_session, name="pt-include", with_resource=True)

    tmpl = (
        make_query(db_session)
        .process_templates()
        .filter(id=run.template.id)
        .include_step_templates()
        .include_resource_slots()
        .first()
    )

    assert tmpl is not None
    assert "Step-pt-include" in tmpl.step_templates
    assert any(rs.name.startswith("slot-pt-include") for rs in tmpl.resource_slots)


def test_resource_queries_can_return_refs(db_session):
    resource_type = ResourceType(name="rt")
    resource_template = ResourceTemplate(name="rtmpl", version="1.0")
    resource_template.types.append(resource_type)
    resource = Resource(name="res-ref", template=resource_template)
    db_session.add_all([resource_type, resource_template, resource])
    db_session.commit()

    res_ref = (
        make_query(db_session).resources(expand=False).filter(id=resource.id).first()
    )
    tmpl_ref = (
        make_query(db_session)
        .resource_templates(expand=False)
        .filter(id=resource_template.id)
        .first()
    )

    assert isinstance(res_ref, ResourceRef)
    assert isinstance(res_ref.template, ResourceTemplateRef)
    assert isinstance(tmpl_ref, ResourceTemplateRef)


def test_resource_template_includes(db_session):
    resource_type = ResourceType(name="rt-inc")
    parent = ResourceTemplate(name="rt-parent", version="1.0")
    parent.types.append(resource_type)

    child = ResourceTemplate(name="rt-child", version="1.0", parent=parent)
    ag = AttributeGroupTemplate(name="Props-inc", resource_template=parent)
    ag.attribute_templates.append(
        AttributeTemplate(name="length", value_type="int", default_value="5")
    )

    db_session.add_all([resource_type, parent, child, ag])
    db_session.commit()

    tmpl = (
        make_query(db_session)
        .resource_templates()
        .filter(id=parent.id)
        .include_children()
        .include_attribute_groups()
        .include_types()
        .first()
    )

    assert tmpl is not None
    assert "rt-child" in tmpl.children
    assert any(
        at.name == "length"
        for at in tmpl.attribute_group_templates[0].attribute_templates
    )
    assert any(t.name == "rt-inc" for t in tmpl.types)



from importlib import resources
from pathlib import Path

from alembic import command
from alembic.config import Config


def get_migration_path() -> Path:
    """Return the filesystem path to the bundled Alembic migrations."""
    return Path(resources.files("recap.db").joinpath("migrations"))


def build_alembic_config(db_url: str, script_location: Path | None = None) -> Config:
    """Create an Alembic Config pointing at the packaged migrations and a DB URL."""
    config = Config()
    location = script_location or get_migration_path()
    config.set_main_option("script_location", str(location))
    config.set_main_option("sqlalchemy.url", db_url)
    return config


def apply_migrations(db_url: str, revision: str = "head") -> None:
    """Apply all migrations up to the requested revision for the given database."""
    config = build_alembic_config(db_url)
    command.upgrade(config, revision)


def downgrade_migrations(db_url: str, revision: str = "base") -> None:
    """Downgrade migrations to the requested revision for the given database."""
    config = build_alembic_config(db_url)
    command.downgrade(config, revision)



from recap.utils.general import Direction, generate_uppercase_alphabets


def test_fragment_screening_api(client):
    # Testing
    # - Create templates
    #     - Library plate
    #     - Xtal plate
    #     - Puck collection
    #     - Puck template
    #     - Pin template
    #     - Process template
    # - Create instance
    #     - Campaign
    #     - ProcessRun
    #     - Library plate
    #     - xtal plate
    #     -

    with client.build_resource_template(
        name="Library Plate 1536",
        type_names=["container", "library_plate", "plate"],
    ) as lp:
        lp.add_properties(
            {
                "LB1536_dimensions": [
                    {"name": "rows", "type": "int", "default": 32},
                    {"name": "columns", "type": "int", "default": 48},
                ]
            }
        )

        a_to_af = generate_uppercase_alphabets(32)
        lib_well_type_names_1536 = [
            {"name": f"{i}{str(j).zfill(2)}"} for i in a_to_af for j in range(1, 49)
        ]
        for well_data in lib_well_type_names_1536:
            lp.add_child(well_data["name"], ["container", "well"]).add_properties(
                {
                    "well_status": [
                        {"name": "used", "type": "bool", "default": False},
                    ],
                    "content": [
                        {"name": "catalog_id", "type": "str", "default": ""},
                        {"name": "SMILES", "type": "str", "default": ""},
                        {"name": "sequence", "type": "int", "default": 0},
                    ],
                }
            ).close_child()

    rt = (
        client.query_maker()
        .resource_templates()
        .filter_by_types(["library_plate"])
        .first()
    )
    assert rt.name == "Library Plate 1536"

    with client.build_resource_template(
        name="SwissCI-MRC-2d", type_names=["container", "xtal_plate", "plate"]
    ) as plate:
        a_to_h = generate_uppercase_alphabets(8)
        a_to_p = generate_uppercase_alphabets(16)
        echo = [f"{i}{j}" for i in a_to_p for j in range(1, 13)]
        shifter = [
            f"{i}{k}{j}" for i in a_to_h for j in ["a", "b"] for k in range(1, 13)
        ]
        plate_maps = [
            {"echo": i, "shifter": j} for i, j in zip(echo, shifter, strict=False)
        ]
        # plate.prop_group("metadata").add_attribute("drop_volume", "int", "nL", 0).close_group()
        plate.add_properties(
            {
                "metadata": [
                    {"name": "drop_volume", "type": "float", "unit": "nL", "default": 0}
                ]
            }
        )
        for plate_map in plate_maps:
            plate_shift_b = plate_map["shifter"][-1] == "b"
            plate.add_child(plate_map["shifter"], ["container", "well"]).add_properties(
                {
                    "echo_offset": [
                        {"name": "x", "type": "int", "default": 0},
                        {
                            "name": "y_0" if plate_shift_b else "y_1350",
                            "type": "int",
                            "default": 0 if plate_shift_b else 1350,
                        },
                        {
                            "name": f"echo_pos_{plate_map['echo']}",
                            "type": "str",
                            "default": plate_map["echo"],
                        },
                    ]
                }
            ).close_child()

    rt = (
        client.query_maker()
        .resource_templates()
        .filter_by_types(["xtal_plate"])
        .first()
    )
    assert rt.name == "SwissCI-MRC-2d"

    with client.build_resource_template(
        name="puck_collection", type_names=["container"]
    ) as puck_collection:
        puck_collection.prop_group("contents").add_attribute("count", "int", "", 0)

    with client.build_resource_template(
        name="puck", type_names=["container", "puck"]
    ) as puck_template:
        puck_template.prop_group("pin_count").add_attribute(
            "total", "int", "", 16
        ).add_attribute("occupied", "int", "", 0).close_group()

    with client.build_resource_template(
        name="pin", type_names=["container", "pin"]
    ) as pin_template:
        pin_template.prop_group("content").add_attribute(
            "position", "int", "", 0
        ).close_group()

    with client.build_process_template("Fragment Screening Sample Prep", "1.0") as pt:
        pt.add_resource_slot(
            "library_plate", "plate", Direction.input
        ).add_resource_slot("xtal_plate", "plate", Direction.input).add_resource_slot(
            "puck_tray", "container", Direction.output
        )
        pt.add_step(name="Image plate").param_group("drop").add_attribute(
            "volume", "float", "nL", 0
        ).close_group().bind_slot("xtal_container", "xtal_plate").close_step()
        pt.add_step(name="Echo transfer").param_group("volume").add_attribute(
            "transferred", "float", "uL", 0
        ).close_group().param_group("batch").add_attribute(
            "number", "int", "", 0
        ).close_group().bind_slot("source_container", "library_plate").bind_slot(
            "dest_container", "xtal_plate"
        ).close_step()

        pt.add_step(name="Harvesting").param_group("harvesting").add_attribute(
            "departure_time", "datetime", "", None
        ).add_attribute("arrival_time", "datetime", "", None).add_attribute(
            "comment", "str", "", ""
        ).add_attribute("status", "str", "", "").close_group().param_group(
            "lsdc"
        ).add_attribute("sample_name", "str", "", "").close_group().bind_slot(
            "source_container", "xtal_plate"
        ).bind_slot("dest_container", "puck_tray").close_step()

    client.create_campaign("Test campaign", "123", "0")

    xtal_plate_template = (
        client.query_maker()
        .resource_templates()
        .filter_by_types(["xtal_plate"])
        .first()
    )
    if xtal_plate_template:
        client.create_resource("Test Xtal plate", xtal_plate_template.name)

    library_plate_template = (
        client.query_maker()
        .resource_templates()
        .filter_by_types(["library_plate"])
        .filter(name="Library Plate 1536")
        .first()
    )
    if library_plate_template:
        with client.build_resource(
            "DSI-poised", library_plate_template.name
        ) as lib_plate_builder:
            for well in lib_plate_builder.resource.children.values():
                well.properties["content"].values["SMILES"] = ""
                well.properties["content"].values["catalog_id"] = ""

    with client.build_process_run(
        "Test run", "Test run for something", "Fragment Screening Sample Prep", "1.0"
    ):
        pass

    ## Testing queries

    qm = client.query_maker()

    campaigns = qm.campaigns()
    runs = qm.process_runs()

    all_campaigns = campaigns.filter(proposal="123").all()
    assert all_campaigns[0].name == "Test campaign"

    assert runs.filter(campaign__proposal="123").count() == 1

    # for run in runs.include_steps(include_parameters=True).all():
    #     # print(f"Run: {run.name}")
    #     for step_num, step in enumerate(run.steps):
    #         # print(f"\tStep {step_num}: {step.name}")
    #         for pg_num, (param_group_name, param_group) in enumerate(
    #             step.parameters.items()
    #         ):
    #             vals = param_group.values
    #             # print(f"\t\tGroup {pg_num}: {param_group_name}")
    #             for param_name, param_value in vals.model_dump(by_alias=True).items():
    #                 pass
    # print(f"\t\t\t{param_name} : {param_value}")

    # for c in campaigns.include_process_runs().all():
    #     print(c.name)
    #     for run in c.process_runs:
    #         print(run.name)

    # plates = resources.filter(template__types__name_in=["library_plate"]).include_template().all()

    # for plate in plates:
    #     print(f"{plate.name} : {plate.template.name}")
